"""The curve commands module.

These commands are used in the following models:
DPO2K, DPO2KB, DPO4K, DPO4KB, DPO5KB, DPO70KC, DPO70KD, DPO70KDX, DPO70KSX, DPO7KC, DSA70KC,
DSA70KD, MDO3, MDO3K, MDO4K, MDO4KB, MDO4KC, MSO2K, MSO2KB, MSO4K, MSO4KB, MSO5KB, MSO70KC, MSO70KDX

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:

::

    - CURVe {<Block>|<asc curve>}
    - CURVe?
"""
from typing import Optional, TYPE_CHECKING

from .._helpers import SCPICmdRead, SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class Curve(SCPICmdWrite, SCPICmdRead):
    """The ``CURVe`` command.

    **Description:**
        - The ``CURVe`` command transfers the waveform data points the oscilloscope's internal
          reference memory location (REF1-4), which is specified by the to ``DATa:DESTination``
          command. The ``CURVe?`` query transfers data the oscilloscope; the source waveform is
          specified by the from ``DATa:SOUrce`` command. The first and last data points are
          specified by the ``DATa:STARt`` and ``DATa:STOP`` commands. Associated with each waveform
          transferred using the ``CURVe`` command or query is a waveform preamble that provides the
          data format, scale and associated information needed to interpret the waveform data
          points. The preamble information for waveforms sent the oscilloscope is specified using
          the to WFMInpre commands. The preamble information for waveforms transferred the
          oscilloscope is specified or queried using the from WFMOutpre commands. If the waveform is
          not displayed, the query form generates an error. The ``CURVe`` command and ``CURVe?``
          query transfer waveform data in ASCII or binary format. ASCII data is sent as a
          comma-separated list of decimal values. Binary data is sent with the IEEE488.2 binary
          block header immediately followed by the binary data. The IEEE488.2 binary block header is
          defined as follows: #N<N-digits> where: N is a single decimal or hexadecimal digit
          indicating the number of digits to follow. <N-digits> are the decimal digits representing
          the number of bytes in the data that immediately follows this binary block header. The
          Waveform Transfer command group text contains more comprehensive information.

    **Usage:**
        - Using the ``.query()`` method will send the ``CURVe?`` query.
        - Using the ``.verify(value)`` method will send the ``CURVe?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``CURVe value`` command.

    **SCPI Syntax:**

    ::

        - CURVe {<Block>|<asc curve>}
        - CURVe?

    **Info:**
        - ``<Block>`` is the waveform data in binary format. The waveform is formatted as follows.
        - ``<asc curve>`` is the waveform data in ASCII format. The format for ASCII data is
          <NR1>[,<NR1>..], where each <NR1> represents a data point. For RF frequency domain
          waveforms, the data is transmitted as 4-byte floating point values (NR2 or NR3).
    """

    def __init__(self, device: Optional["PIDevice"] = None, cmd_syntax: str = "CURVe") -> None:
        super().__init__(device, cmd_syntax)



"""Base AWG device driver module."""
import inspect
import os
import struct

from abc import ABC
from dataclasses import dataclass
from functools import cached_property
from typing import Literal, Type

from tm_devices.driver_mixins.signal_generator_mixin import SourceDeviceConstants
from tm_devices.drivers.device import family_base_class
from tm_devices.drivers.pi.signal_sources.signal_source import SignalSource
from tm_devices.helpers import DeviceTypes, SignalSourceFunctionsAWG


@dataclass(frozen=True)
class AWGSourceDeviceConstants(SourceDeviceConstants):
    """Class to hold source device constants."""

    functions: Type[SignalSourceFunctionsAWG] = SignalSourceFunctionsAWG


@family_base_class
class AWG(SignalSource, ABC):
    """Base AWG device driver."""

    _DEVICE_TYPE = DeviceTypes.AWG.value

    ################################################################################################
    # Magic Methods
    ################################################################################################

    ################################################################################################
    # Properties
    ################################################################################################
    @property
    def source_device_constants(self) -> AWGSourceDeviceConstants:
        """Return the device constants."""
        return self._DEVICE_CONSTANTS  # type: ignore

    @cached_property
    def total_channels(self) -> int:
        """Return the total number of channels (all types)."""
        return int(self.query("AWGControl:CONFigure:CNUMber?", verbose=False))

    ################################################################################################
    # Public Methods
    ################################################################################################
    def load_waveform(self, wfm_name: str, waveform_file_path: str, wfm_type: str) -> None:
        """Load a waveform into the memory of the AWG.

        Args:
            wfm_name: The name to set the loaded waveform to on the AWG.
            waveform_file_path: The path to the waveform.
            wfm_type: The type of the waveform to import.
        """
        if not waveform_file_path.startswith('"'):
            waveform_file_path = '"' + waveform_file_path
        if not waveform_file_path.endswith('"'):
            waveform_file_path += '"'
        self.write(f'MMEMory:IMPort "{wfm_name}", {waveform_file_path}, {wfm_type}')
        self._ieee_cmds.opc()

    def generate_waveform(  # noqa: PLR0913  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        frequency: float,
        function: AWGSourceDeviceConstants,
        amplitude: float,
        offset: float,
        channel: str = "all",
        burst: int = 0,
        termination: Literal["FIFTY", "HIGHZ"] = "FIFTY",
        duty_cycle: float = 50.0,
        polarity: Literal["NORMAL", "INVERTED"] = "NORMAL",
        symmetry: float = 50.0,
    ) -> None:
        """Generate a signal given the following parameters.

        Args:
            frequency: The frequency of the waveform to generate.
            function: The waveform shape to generate.
            amplitude: The amplitude of the signal to generate.
            offset: The offset of the signal to generate.
            channel: The channel name to output the signal from, or 'all'.
            burst: The number of wavelengths to be generated.
            termination: The impedance this device's ``channel`` expects to see at the received end.
            duty_cycle: The duty cycle percentage within [10.0, 90.0].
            polarity: The polarity to set the signal to.
            symmetry: The symmetry to set the signal to, only applicable to certain functions.

        Raises:
            NotImplementedError: Indicates the current driver has not implemented this method.
        """
        # TODO: implement
        raise NotImplementedError(
            f"``.{inspect.currentframe().f_code.co_name}()``"  # pyright: ignore
            f" is not yet implemented for the {self.__class__.__name__} driver"
        )

    ################################################################################################
    # Private Methods
    ################################################################################################
    def _reboot(self) -> None:
        """Reboot the device."""
        # TODO: overwrite the reboot code here

    # TODO: add testing for this
    def _send_waveform(self, target_file: str) -> None:  # pragma: no cover
        """Send the waveform information to the AWG as a file in memory.

        Args:
            target_file: The name of the waveform file.
        """
        with open(target_file, "rb") as file_handle:
            file_handle.seek(0, 0)
            waveform_data = file_handle.read()
            # must be even to send
            if len(waveform_data) % 2:
                waveform_data += b"\0"

            info_len = int(len(waveform_data) / 2)
            bin_waveform = struct.unpack(">" + str(info_len) + "H", waveform_data)

            # Turn "path/to/stuff.wfm" into "stuff.wfm".
            filename_target = os.path.basename(target_file)
            # Write the waveform data to the AWG memory.
            string_to_send = 'MMEMORY:DATA "' + filename_target + '",'
            self._visa_resource.write_binary_values(
                string_to_send,
                bin_waveform,
                datatype="H",
                is_big_endian=True,
                termination="\r\n",
                encoding="latin-1",
            )



"""The application commands module.

These commands are used in the following models:
LPD6, MSO4, MSO5, MSO5B, MSO5LP, MSO6, MSO6B

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:

::

    - APPLication:ACTivate <QString>
"""
from typing import Optional, TYPE_CHECKING

from .._helpers import SCPICmdRead, SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class ApplicationActivate(SCPICmdWrite):
    """The ``APPLication:ACTivate`` command.

    **Description:**
        - For legacy TekExpress applications. Starts the application specified in the string. The
          available applications depend on the oscilloscope model and installed options.

    **Usage:**
        - Using the ``.write(value)`` method will send the ``APPLication:ACTivate value`` command.

    **SCPI Syntax:**

    ::

        - APPLication:ACTivate <QString>

    **Info:**
        - ``<QString>`` is the name of the application that you want to start. Enter the application
          name exactly as it appears in the oscilloscope Applications menu.
    """

    _WRAP_ARG_WITH_QUOTES = True


class Application(SCPICmdRead):
    """The ``APPLication`` command tree.

    **Usage:**
        - Using the ``.query()`` method will send the ``APPLication?`` query.
        - Using the ``.verify(value)`` method will send the ``APPLication?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.activate``: The ``APPLication:ACTivate`` command.
    """

    def __init__(
        self, device: Optional["PIDevice"] = None, cmd_syntax: str = "APPLication"
    ) -> None:
        super().__init__(device, cmd_syntax)
        self._activate = ApplicationActivate(device, f"{self._cmd_syntax}:ACTivate")

    @property
    def activate(self) -> ApplicationActivate:
        """Return the ``APPLication:ACTivate`` command.

        **Description:**
            - For legacy TekExpress applications. Starts the application specified in the string.
              The available applications depend on the oscilloscope model and installed options.

        **Usage:**
            - Using the ``.write(value)`` method will send the ``APPLication:ACTivate value``
              command.

        **SCPI Syntax:**

        ::

            - APPLication:ACTivate <QString>

        **Info:**
            - ``<QString>`` is the name of the application that you want to start. Enter the
              application name exactly as it appears in the oscilloscope Applications menu.
        """
        return self._activate



# pylint: disable=too-many-lines
"""Device manager module."""
from __future__ import annotations

import contextlib
import inspect
import json
import os
import pathlib
import socket
import warnings

from types import FrameType, MappingProxyType, TracebackType
from typing import cast, Dict, Mapping, Optional, Tuple, Type, TYPE_CHECKING, Union

from tm_devices.components import DMConfigParser
from tm_devices.drivers import (
    AFG3K,
    AFG3KB,
    AFG3KC,
    AFG31K,
    AWG5K,
    AWG5KB,
    AWG5KC,
    AWG7K,
    AWG7KB,
    AWG7KC,
    AWG70KA,
    AWG70KB,
    AWG5200,
    DAQ6510,
    DMM6500,
    DMM7510,
    DMM7512,
    DPO2K,
    DPO2KB,
    DPO4K,
    DPO4KB,
    DPO5K,
    DPO5KB,
    DPO7K,
    DPO7KC,
    DPO70K,
    DPO70KC,
    DPO70KD,
    DPO70KDX,
    DPO70KSX,
    DSA70K,
    DSA70KC,
    DSA70KD,
    LPD6,
    MDO3,
    MDO3K,
    MDO4K,
    MDO4KB,
    MDO4KC,
    MSO2,
    MSO2K,
    MSO2KB,
    MSO4,
    MSO4K,
    MSO4KB,
    MSO5,
    MSO5B,
    MSO5K,
    MSO5KB,
    MSO5LP,
    MSO6,
    MSO6B,
    MSO70K,
    MSO70KC,
    MSO70KDX,
    PSU2200,
    PSU2220,
    PSU2230,
    PSU2231,
    PSU2231A,
    PSU2280,
    PSU2281,
    SMU2400,
    SMU2401,
    SMU2410,
    SMU2450,
    SMU2460,
    SMU2461,
    SMU2470,
    SMU2601A,
    SMU2601B,
    SMU2601BPulse,
    SMU2602A,
    SMU2602B,
    SMU2604A,
    SMU2604B,
    SMU2606B,
    SMU2611A,
    SMU2611B,
    SMU2612A,
    SMU2612B,
    SMU2614A,
    SMU2614B,
    SMU2634A,
    SMU2634B,
    SMU2635A,
    SMU2635B,
    SMU2636A,
    SMU2636B,
    SMU2651A,
    SMU2657A,
    SMU6430,
    SMU6514,
    SMU6517B,
    SS3706A,
    TekScopeSW,
    TMT4,
    TSOVu,
)
from tm_devices.drivers.api.rest_api.rest_api_device import RESTAPIDevice
from tm_devices.drivers.pi.pi_device import PIDevice
from tm_devices.helpers import (
    AliasDict,
    check_for_update,
    ConnectionTypes,
    create_visa_connection,
    detect_visa_resource_expression,
    DeviceConfigEntry,
    DeviceTypes,
    DMConfigOptions,
    get_model_series,
    print_with_timestamp,
    PYVISA_PY_BACKEND,
    SerialConfig,
    Singleton,
)

with warnings.catch_warnings():
    warnings.simplefilter("ignore", UserWarning)
    import pyvisa as visa

    from pyvisa_py.protocols.rpc import RPCError  # type: ignore

# noinspection PyUnresolvedReferences  # pylint: disable=unused-import,wrong-import-order
from traceback_with_variables import activate_by_import  # noqa: F401  # type: ignore

if TYPE_CHECKING:
    from pyvisa.resources import MessageBasedResource
    from typing_extensions import Self, TypeAlias

    from tm_devices.drivers.device import Device

####################################################################################################
# Type Aliases
####################################################################################################
# TODO: this is temporary until python3.12 which will support TypeVar with defaults
AFGAlias: TypeAlias = Union[AFG3K, AFG3KB, AFG3KC, AFG31K]
AWGAlias: TypeAlias = Union[AWG5K, AWG5KB, AWG5KC, AWG7K, AWG7KB, AWG7KC, AWG5200, AWG70KA, AWG70KB]
DataAcquisitionSystemAlias: TypeAlias = Union[DAQ6510]  # pyright: ignore
DigitalMultimeterAlias: TypeAlias = Union[DMM6500, DMM7510, DMM7512]
ScopeAlias: TypeAlias = Union[
    DPO5K,
    DPO5KB,
    DPO7K,
    DPO7KC,
    DPO70K,
    DPO70KC,
    DPO70KD,
    DPO70KDX,
    DPO70KSX,
    DSA70K,
    DSA70KC,
    DSA70KD,
    LPD6,
    MSO5,
    MSO6,
    MSO4,
    MSO2,
    MSO6B,
    MSO5B,
    MSO5LP,
    TekScopeSW,
    TSOVu,
    MSO2K,
    MSO2KB,
    DPO2K,
    DPO2KB,
    MDO3,
    MDO3K,
    MDO4K,
    MDO4KB,
    MDO4KC,
    MSO4K,
    MSO4KB,
    DPO4K,
    DPO4KB,
    MSO5K,
    MSO5KB,
    MSO70K,
    MSO70KC,
    MSO70KDX,
]
MarginTesterAlias: TypeAlias = Union[TMT4]  # pyright: ignore
PowerSupplyUnitAlias: TypeAlias = Union[
    PSU2200,
    PSU2220,
    PSU2230,
    PSU2231,
    PSU2231A,
    PSU2280,
    PSU2281,
]
SourceMeasureUnitAlias: TypeAlias = Union[
    SMU2400,
    SMU2401,
    SMU2410,
    SMU2450,
    SMU2460,
    SMU2461,
    SMU2470,
    SMU2601B,
    SMU2601BPulse,
    SMU2602B,
    SMU2604B,
    SMU2606B,
    SMU2611B,
    SMU2612B,
    SMU2614B,
    SMU2634B,
    SMU2635B,
    SMU2636B,
    SMU2651A,
    SMU2657A,
    SMU2601A,
    SMU2602A,
    SMU2604A,
    SMU2611A,
    SMU2612A,
    SMU2614A,
    SMU2634A,
    SMU2635A,
    SMU2636A,
    SMU6430,
    SMU6514,
    SMU6517B,
]
SystemsSwitchAlias: TypeAlias = Union[SS3706A]  # pyright: ignore


####################################################################################################
# DeviceManager class
####################################################################################################
# pylint: disable=too-many-public-methods,too-many-instance-attributes
class DeviceManager(metaclass=Singleton):
    """The manager for all devices.

    Once instantiated, this class creates and manages connections to all devices.

    This class is a singleton, meaning only one instance can ever be created. Any subsequent
    instantiation attempts will return a pointer to the original instance.
    """

    ################################################################################################
    # Magic Methods
    ################################################################################################
    def __init__(
        self,
        verbose: bool = True,
        config_options: Optional[DMConfigOptions] = None,
        external_device_drivers: Optional[Mapping[str, Type[Device]]] = None,
    ) -> None:
        """Create the instance of the DeviceManager.

        Args:
            verbose: A boolean indicating if verbose output should be printed.
            config_options: An optional set of configuration options to use
                (updates the current configuration options).
            external_device_drivers: An optional dict for passing in additional device drivers.
        """
        check_for_update()
        self._suppress_protection = False
        # Set up the DeviceManager
        self.__is_open = False
        self.__verbose_visa = False
        self.__devices: Dict[str, Union[PIDevice, RESTAPIDevice]] = AliasDict()
        self._external_device_drivers = external_device_drivers
        self.__config = DMConfigParser()
        if config_options is not None:
            self.__config.update_options(config_options)

        # initialize for __set_options
        self.__verbose: bool = NotImplemented
        self.__teardown_cleanup_enabled: bool = NotImplemented
        self.__setup_cleanup_enabled: bool = NotImplemented
        self.__visa_library: str = NotImplemented
        # actually populate the options
        self.__set_options(verbose)

        self.open()

    def __del__(self) -> None:
        """Delete the instance of the DeviceManger."""
        if self.__is_open:
            self.close()

    def __enter__(self) -> Self:
        """Provide access to the DeviceManager class instance.

        This and the __exit__ method allow this class to be used as a Context Manager.
        """
        if not self.__is_open:
            self.open()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Handle exiting the context by closing down the DeviceManager.

        This and the __enter__ method allow this class to be used as a Context Manager.
        """
        if self.__is_open:
            self.close()

    ################################################################################################
    # Properties
    ################################################################################################
    @property
    def devices(self) -> Mapping[str, Union[RESTAPIDevice, PIDevice]]:
        """Return the dictionary of devices."""
        return MappingProxyType(self.__devices)

    @property
    def is_open(self) -> bool:
        """Return a boolean indicating if the DeviceManager has closed all device connections."""
        return self.__is_open

    @property
    def teardown_cleanup_enabled(self) -> bool:
        """Return a boolean indicating if cleanup at teardown is enabled for devices."""
        return self.__teardown_cleanup_enabled

    @teardown_cleanup_enabled.setter
    def teardown_cleanup_enabled(self, value: bool) -> None:
        """Set the property which can enable cleanup on teardown of devices."""
        self.__config.options.teardown_cleanup = value
        self.__teardown_cleanup_enabled = value

    @property
    def setup_cleanup_enabled(self) -> bool:
        """Return a boolean indicating if cleanup at setup is enabled for devices."""
        return self.__setup_cleanup_enabled

    @setup_cleanup_enabled.setter
    def setup_cleanup_enabled(self, value: bool) -> None:
        """Set the property which can enable cleanup on setup of devices."""
        self.__config.options.setup_cleanup = value
        self.__setup_cleanup_enabled = value

    @property
    def verbose(self) -> bool:
        """Return the verbosity setting of the DeviceManager."""
        return self.__verbose

    @verbose.setter
    def verbose(self, value: bool) -> None:
        """Set the verbosity of the DeviceManager."""
        self.__config.options.verbose_mode = value
        self.__verbose = value

    @property
    def verbose_visa(self) -> bool:
        """Return the VISA verbosity setting of the DeviceManager."""
        return self.__verbose_visa

    @verbose_visa.setter
    def verbose_visa(self, value: bool) -> None:
        """Set the VISA verbosity setting of the DeviceManager."""
        if value:
            visa.log_to_screen()
        else:
            for handler in visa.logger.handlers.copy():
                if "DEBUG" in str(handler):
                    visa.logger.removeHandler(handler)
        self.__config.options.verbose_visa = value
        self.__verbose_visa = value

    @property
    def visa_library(self) -> str:
        """Return the currently selected VISA library."""
        return self.__visa_library

    @visa_library.setter
    def visa_library(self, value: str) -> None:
        """Set the VISA library to use when creating new device connections."""
        self.__config.options.standalone = value == PYVISA_PY_BACKEND
        self.__visa_library = value

    ################################################################################################
    # Public Methods
    ################################################################################################
    def add_afg(
        self,
        address: str,
        *,
        alias: Optional[str] = None,
        connection_type: Optional[str] = None,
        port: Optional[int] = None,
        serial_config: Optional[SerialConfig] = None,
    ) -> AFGAlias:
        """Add an AFG to the DeviceManager.

        Args:
            address: The address of the AFG, either an IP address or hostname. If the connection
                     type is ``"USB"`` then the address must be specified as ``"<model>-<serial>"``.
            alias: An optional alias to use to refer to the AFG. If no alias is provided,
                   the device type and number can be used to access the AFG instead.
            connection_type: The type of connection to use for VISA, defaults to TCPIP, not needed
                when the address is a visa resource expression since the connection type is parsed
                from the address string.
            port: The port to use when creating a socket connection.
            serial_config: Serial connection settings, only needed when connection_type="SERIAL".

        Returns:
            The AFG device driver.
        """
        self.__protect_access()
        return cast(
            AFGAlias,
            self._add_device(
                device_type=DeviceTypes.AFG.value,
                address=address,
                alias=alias,
                connection_type=connection_type,
                port=port,
                serial_config=serial_config,
            ),
        )

    def add_awg(
        self,
        address: str,
        *,
        alias: Optional[str] = None,
        connection_type: Optional[str] = None,
        port: Optional[int] = None,
        serial_config: Optional[SerialConfig] = None,
    ) -> AWGAlias:
        """Add an AWG to the DeviceManager.

        Args:
            address: The address of the AWG, either an IP address or hostname. If the connection
                     type is ``"USB"`` then the address must be specified as ``"<model>-<serial>"``.
            alias: An optional alias to use to refer to the AWG. If no alias is provided,
                   the device type and number can be used to access the AWG instead.
            connection_type: The type of connection to use for VISA, defaults to TCPIP, not needed
                when the address is a visa resource expression since the connection type is parsed
                from the address string.
            port: The port to use when creating a socket connection.
            serial_config: Serial connection settings, only needed when connection_type="SERIAL".

        Returns:
            The AWG device driver.
        """
        self.__protect_access()
        return cast(
            AWGAlias,
            self._add_device(
                device_type=DeviceTypes.AWG.value,
                address=address,
                alias=alias,
                connection_type=connection_type,
                port=port,
                serial_config=serial_config,
            ),
        )

    def add_daq(
        self,
        address: str,
        *,
        alias: Optional[str] = None,
        connection_type: Optional[str] = None,
        port: Optional[int] = None,
        serial_config: Optional[SerialConfig] = None,
    ) -> DataAcquisitionSystemAlias:
        """Add a DAQ to the DeviceManager.

        Args:
            address: The address of the DAQ, either an IP address or hostname. If the connection
                     type is ``"USB"`` then the address must be specified as ``"<model>-<serial>"``.
            alias: An optional alias to use to refer to the DAQ. If no alias is provided,
                   the device type and number can be used to access the DAQ instead.
            connection_type: The type of connection to use for VISA, defaults to TCPIP, not needed
                when the address is a visa resource expression since the connection type is parsed
                from the address string.
            port: The port to use when creating a socket connection.
            serial_config: Serial connection settings, only needed when connection_type="SERIAL".

        Returns:
            The DAQ device driver.
        """
        self.__protect_access()
        return cast(
            DataAcquisitionSystemAlias,
            self._add_device(
                device_type=DeviceTypes.DAQ.value,
                address=address,
                alias=alias,
                connection_type=connection_type,
                port=port,
                serial_config=serial_config,
            ),
        )

    def add_dmm(
        self,
        address: str,
        *,
        alias: Optional[str] = None,
        connection_type: Optional[str] = None,
        port: Optional[int] = None,
        serial_config: Optional[SerialConfig] = None,
    ) -> DigitalMultimeterAlias:
        """Add a DMM to the DeviceManager.

        Args:
            address: The address of the DMM, either an IP address or hostname. If the connection
                     type is ``"USB"`` then the address must be specified as ``"<model>-<serial>"``.
            alias: An optional alias to use to refer to the DMM. If no alias is provided,
                   the device type and number can be used to access the AWG instead.
            connection_type: The type of connection to use for VISA, defaults to TCPIP, not needed
                when the address is a visa resource expression since the connection type is parsed
                from the address string.
            port: The port to use when creating a socket connection.
            serial_config: Serial connection settings, only needed when connection_type="SERIAL".

        Returns:
            The DMM device driver.
        """
        self.__protect_access()
        return cast(
            DigitalMultimeterAlias,
            self._add_device(
                device_type=DeviceTypes.DMM.value,
                address=address,
                alias=alias,
                connection_type=connection_type,
                port=port,
                serial_config=serial_config,
            ),
        )

    def add_mt(
        self,
        address: str,
        device_driver: str,
        *,
        alias: Optional[str] = None,
        port: Optional[int] = None,
    ) -> MarginTesterAlias:
        """Add a Margin Tester to the DeviceManager.

        Args:
            address: The address of the Margin Tester, either an IP address or hostname.
            device_driver: A string indicating the specific Python device driver to use.
            alias: An optional alias to use to refer to the Margin Tester. If no alias is provided,
                   the device type and number can be used to access the Margin Tester instead.
            port: The port to use when creating the connection.

        Returns:
            The Margin Tester device driver.
        """
        self.__protect_access()
        return cast(
            MarginTesterAlias,
            self._add_device(
                device_type=DeviceTypes.MT.value,
                address=address,
                device_driver=device_driver,
                alias=alias,
                connection_type=ConnectionTypes.REST_API.value,
                port=port,
            ),
        )

    def add_psu(
        self,
        address: str,
        *,
        alias: Optional[str] = None,
        connection_type: Optional[str] = None,
        port: Optional[int] = None,
        serial_config: Optional[SerialConfig] = None,
    ) -> PowerSupplyUnitAlias:
        """Add a PSU to the DeviceManager.

        Args:
            address: The address of the PSU, either an IP address or hostname. If the connection
                     type is ``"USB"`` then the address must be specified as ``"<model>-<serial>"``.
            alias: An optional alias to use to refer to the PSU. If no alias is provided,
                   the device type and number can be used to access the PSU instead.
            connection_type: The type of connection to use for VISA, defaults to TCPIP, not needed
                when the address is a visa resource expression since the connection type is parsed
                from the address string.
            port: The port to use when creating a socket connection.
            serial_config: Serial connection settings, only needed when connection_type="SERIAL".

        Returns:
            The PSU device driver.
        """
        self.__protect_access()
        return cast(
            PowerSupplyUnitAlias,
            self._add_device(
                device_type=DeviceTypes.PSU.value,
                address=address,
                alias=alias,
                connection_type=connection_type,
                port=port,
                serial_config=serial_config,
            ),
        )

    def add_scope(
        self,
        address: str,
        *,
        alias: Optional[str] = None,
        connection_type: Optional[str] = None,
        port: Optional[int] = None,
        serial_config: Optional[SerialConfig] = None,
    ) -> ScopeAlias:
        """Add a scope to the DeviceManager.

        Args:
            address: The address of the scope, either an IP address or hostname. If the connection
                     type is ``"USB"`` then the address must be specified as ``"<model>-<serial>"``.
            alias: An optional alias to use to refer to the scope. If no alias is provided,
                   the device type and number can be used to access the scope instead.
            connection_type: The type of connection to use for VISA, defaults to TCPIP, not needed
                when the address is a visa resource expression since the connection type is parsed
                from the address string.
            port: The port to use when creating a socket connection.
            serial_config: Serial connection settings, only needed when connection_type="SERIAL".

        Returns:
            The scope device driver.
        """
        self.__protect_access()
        return cast(
            ScopeAlias,
            self._add_device(
                device_type=DeviceTypes.SCOPE.value,
                address=address,
                alias=alias,
                connection_type=connection_type,
                port=port,
                serial_config=serial_config,
            ),
        )

    def add_smu(
        self,
        address: str,
        *,
        alias: Optional[str] = None,
        connection_type: Optional[str] = None,
        port: Optional[int] = None,
        serial_config: Optional[SerialConfig] = None,
    ) -> SourceMeasureUnitAlias:
        """Add a SMU to the DeviceManager.

        Args:
            address: The address of the device, either an IP address or hostname. If the connection
                     type is ``"USB"`` then the address must be specified as ``"<model>-<serial>"``.
            alias: An optional alias to use to refer to the SMU. If no alias is provided,
                   the device type and number can be used to access the SMU instead.
            connection_type: The type of connection to use for VISA, defaults to TCPIP, not needed
                when the address is a visa resource expression since the connection type is parsed
                from the address string.
            port: The port to use when creating a socket connection.
            serial_config: Serial connection settings, only needed when connection_type="SERIAL".

        Returns:
            The SMU device driver.
        """
        self.__protect_access()
        return cast(
            SourceMeasureUnitAlias,
            self._add_device(
                device_type=DeviceTypes.SMU.value,
                address=address,
                alias=alias,
                connection_type=connection_type,
                port=port,
                serial_config=serial_config,
            ),
        )

    def add_ss(
        self,
        address: str,
        *,
        alias: Optional[str] = None,
        connection_type: Optional[str] = None,
        port: Optional[int] = None,
    ) -> SystemsSwitchAlias:
        """Add a SS to the DeviceManager.

        Args:
            address: The address of the device, either an IP address or hostname. If the connection
                     type is ``"USB"`` then the address must be specified as ``"<model>-<serial>"``.
            alias: An optional alias to use to refer to the SS. If no alias is provided,
                   the device type and number can be used to access the SS instead.
            connection_type: The type of connection to use for VISA, defaults to TCPIP, not needed
            when the address is a visa resource expression since the connection type is parsed from
            the address string.
            port: The port to use when creating a socket connection.

        Returns:
            The SS device driver.
        """
        self.__protect_access()
        return cast(
            SystemsSwitchAlias,
            self._add_device(
                device_type=DeviceTypes.SS.value,
                address=address,
                alias=alias,
                connection_type=connection_type,
                port=port,
            ),
        )

    def cleanup_all_devices(self) -> None:
        """Cleanup and reset all devices."""
        self.__protect_access()
        for device_name, device_object in self.__devices.items():
            try:
                device_object.cleanup(verbose=False or bool(self.__config.options.verbose_mode))
            except (visa.errors.Error, socket.error, RPCError, AttributeError):  # noqa: PERF203
                print(f"Device cleanup of {device_name} failed. Retrying...")
                device_object.cleanup()

    def close(self) -> None:
        """Close the DeviceManager."""
        self.__protect_access()
        print("")
        if self.__devices:
            print_with_timestamp("Closing Connections to Devices")
            if self.__teardown_cleanup_enabled:
                self.cleanup_all_devices()
            for device_object in list(self.__devices.values()):
                device_object.close()
        self.__is_open = False
        print_with_timestamp(f"{self.__class__.__name__} Closed")

    def disable_device_command_checking(self) -> None:
        """Set the `.enable_verification` attribute of each device to `False`.

        This has the effect of speeding up automation scripts that no longer require checking each
        command after it is sent.
        """
        self.__protect_access()
        for device in self.__devices.values():
            device.enable_verification = False

    def get_available_devices(
        self, search: str = "", configured: bool = True, local: bool = True
    ) -> Dict[str, Tuple[str, ...]]:
        """Get tuples of local and configured devices, optionally narrowed by a search.

        Args:
            search: Simple case-insensitive search criteria for available devices.
            configured: Whether to include configured devices.
            local: Whether to include local devices.

        Returns:
            A dictionary of found devices.

        Raises:
            AssertionError: Local resources can't be queried with the VISA ResourceManager.
                May need to configure the VISA backend.
        """
        self.__protect_access()
        found_devices: Dict[str, Tuple[str, ...]] = {}
        if configured:
            found_devices["configured"] = tuple(
                str(device_config) for device_config in self.__config.devices.values()
            )
        if local:
            try:
                resource_manager = visa.ResourceManager(self.__visa_library)
                found_devices["local"] = resource_manager.list_resources()
            except visa.Error as exc:
                msg = (
                    "Unable to query local resources, ensure you have a suitable "
                    "VISA backend properly installed (NI-VISA is recommended) or "
                    "specify STANDALONE in your config file."
                )
                raise AssertionError(msg) from exc
        if search:
            if configured:
                found_devices["configured"] = tuple(
                    x for x in found_devices["configured"] if search.upper() in x.upper()
                )
            if local:
                found_devices["local"] = tuple(
                    x for x in found_devices["local"] if search.upper() in x.upper()
                )

        return found_devices

    def get_afg(self, number_or_alias: Union[int, str]) -> AFGAlias:
        """Get the AFG driver for the given AFG number or alias.

        Integers are treated as a device number, strings are treated as an alias.

        Args:
            number_or_alias: The number or alias of the AFG to get.

        Returns:
            The AFG device driver.
        """
        self.__protect_access()
        if isinstance(number_or_alias, int):
            return cast(
                AFGAlias,
                self.get_device(device_type=DeviceTypes.AFG.value, device_number=number_or_alias),
            )
        return cast(
            AFGAlias, self.get_device(device_type=DeviceTypes.AFG.value, alias=number_or_alias)
        )

    def get_awg(self, number_or_alias: Union[int, str]) -> AWGAlias:
        """Get the AWG driver for the given AWG number or alias.

        Integers are treated as a device number, strings are treated as an alias.

        Args:
            number_or_alias: The number or alias of the AWG to get.

        Returns:
            The AWG device driver.
        """
        self.__protect_access()
        if isinstance(number_or_alias, int):
            return cast(
                AWGAlias,
                self.get_device(device_type=DeviceTypes.AWG.value, device_number=number_or_alias),
            )
        return cast(
            AWGAlias, self.get_device(device_type=DeviceTypes.AWG.value, alias=number_or_alias)
        )

    def get_daq(self, number_or_alias: Union[int, str]) -> DataAcquisitionSystemAlias:
        """Get the DAQ driver for the given scope number or alias.

        Integers are treated as a device number, strings are treated as an alias.

        Args:
            number_or_alias: The number or alias of the scope to get.

        Returns:
            The scope device driver.
        """
        self.__protect_access()
        if isinstance(number_or_alias, int):
            return cast(
                DataAcquisitionSystemAlias,
                self.get_device(device_type=DeviceTypes.DAQ.value, device_number=number_or_alias),
            )
        return cast(
            DataAcquisitionSystemAlias,
            self.get_device(device_type=DeviceTypes.DAQ.value, alias=number_or_alias),
        )

    def get_dmm(self, number_or_alias: Union[int, str]) -> DigitalMultimeterAlias:
        """Get the DMM driver for the given DMM number or alias.

        Integers are treated as a device number, strings are treated as an alias.

        Args:
            number_or_alias: The number or alias of the DMM to get.

        Returns:
            The DMM device driver.
        """
        self.__protect_access()
        if isinstance(number_or_alias, int):
            return cast(
                DigitalMultimeterAlias,
                self.get_device(device_type=DeviceTypes.DMM.value, device_number=number_or_alias),
            )
        return cast(
            DigitalMultimeterAlias,
            self.get_device(device_type=DeviceTypes.DMM.value, alias=number_or_alias),
        )

    def get_device(
        self,
        *,
        device_type: Optional[str] = None,
        device_number: Optional[Union[int, str]] = None,
        alias: Optional[str] = None,
    ) -> Union[RESTAPIDevice, PIDevice]:
        """Get the driver for the given device.

        Either `device_type` and `device_number` or `alias` must be provided when using this method.

        Args:
            device_type: The type of the device to get.
            device_number: The number of the device to get.
            alias: The alias of the device to get.

        Returns:
            The device driver.

        Raises:
            LookupError: The device was not found in the device driver dictionary.
            ValueError: Indicates that both the type/number and alias parameter were left empty.
            TypeError: Indicates that the fetched device type does not match the specified type.
        """
        if not ((device_type and device_number) or alias):
            message = "Either a name or alias must be specified when fetching a device."
            raise ValueError(message)

        device_name = alias.upper() if alias else f"{device_type} {device_number}".upper()
        try:
            device = self.__devices[device_name]
        except KeyError as error:
            message = f"{device_name} was not found in the device driver dictionary."
            raise LookupError(message) from error
        # double check that the device is the correct type
        if device_type is not None and device.device_type != device_type.upper():
            message = (
                f'A device of type "{device_type}" was specified to be accessed, '
                f'but the accessed device type is actually of type "{device.device_type}".'
            )
            raise TypeError(message)

        return device

    def get_mt(self, number_or_alias: Union[int, str]) -> MarginTesterAlias:
        """Get the Margin Tester driver for the given Margin Tester number or alias.

        Integers are treated as a device number, strings are treated as an alias.

        Args:
            number_or_alias: The number or alias of the Margin Tester to get.

        Returns:
            The Margin Tester device driver.
        """
        self.__protect_access()
        if isinstance(number_or_alias, int):
            return cast(
                MarginTesterAlias,
                self.get_device(device_type=DeviceTypes.MT.value, device_number=number_or_alias),
            )
        return cast(
            MarginTesterAlias,
            self.get_device(device_type=DeviceTypes.MT.value, alias=number_or_alias),
        )

    def get_psu(self, number_or_alias: Union[int, str]) -> PowerSupplyUnitAlias:
        """Get the PSU driver for the given PSU number or alias.

        Integers are treated as a device number, strings are treated as an alias.

        Args:
            number_or_alias: The number or alias of the PSU to get.

        Returns:
            The PSU device driver.
        """
        self.__protect_access()
        if isinstance(number_or_alias, int):
            return cast(
                PowerSupplyUnitAlias,
                self.get_device(device_type=DeviceTypes.PSU.value, device_number=number_or_alias),
            )
        return cast(
            PowerSupplyUnitAlias,
            self.get_device(device_type=DeviceTypes.PSU.value, alias=number_or_alias),
        )

    def get_scope(self, number_or_alias: Union[int, str]) -> ScopeAlias:
        """Get the scope driver for the given scope number or alias.

        Integers are treated as a device number, strings are treated as an alias.

        Args:
            number_or_alias: The number or alias of the scope to get.

        Returns:
            The scope device driver.
        """
        self.__protect_access()
        if isinstance(number_or_alias, int):
            return cast(
                ScopeAlias,
                self.get_device(device_type=DeviceTypes.SCOPE.value, device_number=number_or_alias),
            )
        return cast(
            ScopeAlias, self.get_device(device_type=DeviceTypes.SCOPE.value, alias=number_or_alias)
        )

    def get_smu(self, number_or_alias: Union[int, str]) -> SourceMeasureUnitAlias:
        """Get the SMU driver for the given SMU number or alias.

        Integers are treated as a device number, strings are treated as an alias.

        Args:
            number_or_alias: The number or alias of the SMU to get.

        Returns:
            The SMU device driver.
        """
        self.__protect_access()
        if isinstance(number_or_alias, int):
            return cast(
                SourceMeasureUnitAlias,
                self.get_device(device_type=DeviceTypes.SMU.value, device_number=number_or_alias),
            )
        return cast(
            SourceMeasureUnitAlias,
            self.get_device(device_type=DeviceTypes.SMU.value, alias=number_or_alias),
        )

    def get_ss(self, number_or_alias: Union[int, str]) -> SystemsSwitchAlias:
        """Get the SS driver for the given SS number or alias.

        Integers are treated as a device number, strings are treated as an alias.

        Args:
            number_or_alias: The number or alias of the SS to get.

        Returns:
            The SS device driver.
        """
        self.__protect_access()
        if isinstance(number_or_alias, int):
            return cast(
                SystemsSwitchAlias,
                self.get_device(device_type=DeviceTypes.SS.value, device_number=number_or_alias),
            )
        return cast(
            SystemsSwitchAlias,
            self.get_device(device_type=DeviceTypes.SS.value, alias=number_or_alias),
        )

    def load_config_file(self, config_file_path: Union[str, os.PathLike[str]]) -> None:
        """Load in the config file located at the given path.

        This method will update the current configuration options with any newly defined options and
        add any newly defined devices.

        Args:
            config_file_path: The path to the config file to load.
        """
        print_with_timestamp(
            f"Loading Configuration from {pathlib.Path(config_file_path).resolve()}"
        )
        self.__config.load_config_file(config_file_path)
        self.__set_options(self.__verbose)
        if len(self.__config.devices) != len(self.__devices):
            print_with_timestamp("Opening Connections to Devices")
            for device_name, device_config in self.__config.devices.items():
                if device_name not in self.__devices:
                    self.__create_device(device_name, device_config)

    def open(self) -> bool:  # noqa: A003
        """Reopen all devices if the DeviceManager has been previously closed.

        Returns:
            A boolean indicating if the DeviceManager was opened.
        """
        self._suppress_protection = True
        if self.__is_open:
            msg = (
                f"The .open() method should only be used if the "
                f"{self.__class__.__name__} has already been closed."
            )
            raise AssertionError(msg)
        print_with_timestamp(f"Opening {self.__class__.__name__}")
        # Create the devices
        if self.__config.devices:
            print_with_timestamp("Opening Connections to Devices")
        for device_name, device_config in self.__config.devices.items():
            self.__create_device(device_name, device_config)
        if self.__setup_cleanup_enabled:
            self.cleanup_all_devices()
        self.__is_open = True
        self._suppress_protection = False

        return self.__is_open

    def remove_all_devices(self) -> None:
        """Remove all devices from the DeviceManager."""
        self.__protect_access()
        for device in list(self.__devices.values()):
            self.remove_device(device_type=device.device_type, device_number=device.device_number)

    def remove_device(
        self,
        *,
        device_type: Optional[str] = None,
        device_number: Optional[Union[int, str]] = None,
        alias: Optional[str] = None,
    ) -> None:
        """Remove a device from the DeviceManager.

        Either `device_type` and `device_number` or `alias` must be provided when using this method.

        Args:
            device_type: The type of the device to remove, e.g. 'SCOPE'.
            device_number: The number of the device to remove, e.g. '1'.
            alias: A device alias.
        """
        self.__protect_access()
        # Get the device
        if alias:
            device = self.get_device(alias=alias)
        else:
            device = self.get_device(device_type=device_type, device_number=device_number)
        # cannot depend on device.name here in case of mismatched device_type and device.device_type
        device_name = f"{device.config_entry.device_type.value} {device.device_number}"
        # Close the device
        device.close()
        # Remove the device from the dictionary
        del self.__devices[device_name]
        # Remove the device from the config
        self.__config.remove_device(device_name)

    def write_current_configuration_to_config_file(
        self,
        config_file_path: Optional[Union[str, os.PathLike[str]]] = None,
    ) -> None:
        """Write a config file located at the current working directory (or custom path).

        This method will overwrite any existing config file with the current devices and options.
        This method also respects the ``TM_DEVICES_CONFIG`` environment variable.

        Args:
            config_file_path: The path to the config file. If ends in ".toml" will create toml file.
        """
        print_with_timestamp("Writing Configuration to file")
        new_file_path = pathlib.Path(self.__config.write_config_to_file(config_file_path)).resolve()
        print_with_timestamp(f"Wrote Configuration to {new_file_path}")

    def get_current_configuration_as_environment_variable_strings(self) -> str:
        """Return the current configuration represented as environment variables."""
        return str(self.__config)

    ################################################################################################
    # Private Methods
    ################################################################################################
    def _add_device(  # noqa: PLR0913
        self,
        device_type: str,
        address: str,
        *,
        alias: Optional[str] = None,
        connection_type: Optional[str] = None,
        port: Optional[int] = None,
        serial_config: Optional[SerialConfig] = None,
        device_driver: Optional[str] = None,
    ) -> Union[RESTAPIDevice, PIDevice]:
        """Add a device to the DeviceManager.

        Args:
            device_type: The type of device to add, e.g. 'SCOPE', 'AFG', etc.
            address: The address of the device, either an IP address or hostname. If the connection
                     type is ``"USB"`` then the address must be specified as ``"<model>-<serial>"``.
            alias: An optional alias to use to refer to the device. If no alias is provided,
                   the device type and number can be used to access the device instead.
            connection_type: The type of connection to use for VISA, defaults to TCPIP, not needed
                when the address is a visa resource expression since the connection type is parsed
                from the address string.
            port: The port to use when creating a socket connection.
            serial_config: Serial connection settings, only needed when connection_type="SERIAL".
            device_driver: A string indicating the specific Python device driver to use.

        Returns:
            The created device.
        """
        if (visa_resource_parts := detect_visa_resource_expression(address)) is not None:
            connection_type = visa_resource_parts[0]
            address = visa_resource_parts[1]

        # Device Manager uses all caps for key mappings to device drivers and aliases
        config_dict: dict[str, Optional[Union[str, int, SerialConfig]]] = {
            "device_type": device_type.upper(),
            "address": address.upper(),
        }
        if alias:
            config_dict["alias"] = alias.upper()
        if connection_type:
            config_dict["connection_type"] = connection_type.upper()
        if port:
            config_dict["lan_port"] = port
        if serial_config:
            config_dict["serial_config"] = serial_config
        if device_driver:
            config_dict["device_driver"] = device_driver
        new_device_name, new_device_config = self.__config.add_device(**config_dict)  # type: ignore

        return self.__create_device(new_device_name, new_device_config)

    @staticmethod
    def __clear_visa_output_buffer_and_get_idn(visa_resource: MessageBasedResource) -> str:
        """Clear out the VISA buffer of a device and return the ``*IDN?`` response.

        Args:
            visa_resource: The VISA resource object.

        Returns:
            The response from the ``*IDN?`` query.

        Raises:
            SystemError: Indicates that the buffer was unable to be cleared.
        """
        if 16 & visa_resource.read_stb():
            # 16 is the MAV bit (Message Available) from the Status Byte register
            # MAV flag is only one bit and turns off after a single response is successfully read,
            # even if there's more in the buffer
            warnings.warn(
                f"\nThe device `{visa_resource.resource_info.resource_name}` had data "
                "sitting in the VISA Output Buffer on first connection. "
                "\nDetected data in the buffer via the Status Byte register. "
                "\nThe device_clear() will be called so VISA I/O buffers get flushed.",
                stacklevel=1,
            )
        # always flush the VISA I/O Buffers on the device to clean up any stale data.
        # (note: the Events are kept in different buffers, so *ESR? is not impacted)
        visa_resource.clear()
        visa_resource.write("*IDN?")
        idn_response = ""
        error_msg = None
        old_timeout = visa_resource.timeout
        # 2 second timeout to read should be good
        visa_resource.timeout = 2000
        try:
            msg_string = visa_resource.read().strip()
            # IDN should be a 4 fields separated by 3 commas, also double check string is consistent
            assert msg_string.count(",") == 3  # noqa: S101,PLR2004
            assert msg_string == visa_resource.query("*IDN?").strip()  # noqa: S101
            idn_response = msg_string
        except (UnicodeDecodeError, AssertionError):
            # Unreadable or mismatched response means something is really wrong.
            error_msg = (
                f"Unable to read *IDN? response.\n\t"
                f"The device `{visa_resource.resource_info.resource_name}` likely has data sitting "
                f"in the VISA Output Buffer that survived device_clear()."
            )
        except visa.VisaIOError as error:
            # A timeout means something has gone terribly wrong.
            error_msg = (
                f"{error!s}\n\t"
                "Unable to read data after 2s.\n\t\n\t"
                "Please reboot or read/clear the VISA output buffer on your device and try again."
            )
        if error_msg is not None:
            visa_resource.close()
            raise SystemError(error_msg)
        visa_resource.timeout = old_timeout
        return idn_response

    def __create_device(
        self, device_config_name: str, device_config: DeviceConfigEntry
    ) -> Union[RESTAPIDevice, PIDevice]:
        """Create a new device driver and add it to the device dictionary.

        Args:
            device_config_name: The name returned when creating the device_config.
            device_config: The dataclass holding the device configuration information.

        Returns:
            The created device driver.

        Raises:
            TypeError: Indicates that the created device type does not match the specified type.
            LookupError: Indicates something went wrong when creating the device.
            SystemError: Indicates something went wrong when creating the device.
            AssertionError: Indicates something went wrong when creating the device.
        """
        # pylint: disable=import-outside-toplevel
        from tm_devices.drivers import DEVICE_DRIVER_MODEL_MAPPING

        if self._external_device_drivers is not None:
            device_drivers: Mapping[str, Type[Device]] = MappingProxyType(
                {**self._external_device_drivers, **DEVICE_DRIVER_MODEL_MAPPING}
            )
        else:
            device_drivers = DEVICE_DRIVER_MODEL_MAPPING

        alias_string = f' "{device_config.alias}"' if device_config.alias else ""
        print_with_timestamp(f"Creating Connection to {device_config_name}{alias_string}")
        new_device: Union[RESTAPIDevice, PIDevice]
        if device_config.connection_type == ConnectionTypes.REST_API:
            device_driver_class = device_drivers[str(device_config.device_driver)]
            new_device = cast(RESTAPIDevice, device_driver_class(device_config, self.__verbose))
        else:
            # Create VISA connection and determine proper device driver
            try:
                visa_resource = create_visa_connection(
                    device_config,
                    self.__visa_library,
                    retry_connection=bool(self.__config.options.retry_visa_connection),
                )
                new_device = self.__select_visa_device_driver(
                    visa_resource, device_config, device_drivers
                )
            except (LookupError, SystemError, AssertionError, NotImplementedError, ConnectionError):
                # Catch errors here and remove the device to allow future connections to happen.
                self.__config.remove_device(device_config_name)
                # Simply re-raise the exception, it will have plenty of information on the error.
                raise

        # hardcode the device number from the config entry key name
        new_device._device_number = int(  # noqa: SLF001  # pyright: ignore [reportPrivateUsage]
            device_config_name.rsplit(" ", 1)[1]
        )

        if device_config_name in self.__devices:
            del self.__devices[device_config_name]
        self.__devices[device_config_name] = new_device
        if device_config.alias:
            self.__devices[device_config.alias] = new_device

        # double check created device is correct type
        if new_device.device_type != new_device.config_entry.device_type.value:
            self.remove_device(
                alias=new_device.config_entry.alias,
                device_type=new_device.config_entry.device_type.value,
                device_number=new_device.device_number,
            )
            msg = (
                f'A device of type "{new_device.config_entry.device_type.value}" was specified to '
                f"be created, but the created device type is actually of type "
                f'"{new_device.device_type}".'
            )
            raise TypeError(msg)

        # Clean up devices if the option is enabled.
        if self.__setup_cleanup_enabled:
            new_device.cleanup(verbose=False or bool(self.__config.options.verbose_mode))

        return new_device

    def __protect_access(self) -> None:
        """Check if the method should be allowed to be used.

        Raises:
            AttributeError: Indicates that the calling method should not have been used.
        """
        if not self.__is_open and not self._suppress_protection:
            previous_frame = cast(FrameType, inspect.currentframe().f_back.f_back)  # type: ignore
            message = (
                f"The {self.__class__.__name__} is closed, please use the .open() "
                f"method before continuing to use the {self.__class__.__name__}.\n"
            )
            with contextlib.suppress(AttributeError):
                message += (
                    f"Called from {previous_frame.f_code.co_filename}:"
                    f"{previous_frame.f_lineno}\n"
                )
            raise AttributeError(message)

    def __select_visa_device_driver(
        self,
        visa_resource: MessageBasedResource,
        device_config: DeviceConfigEntry,
        device_drivers: Mapping[str, Type[Device]],
    ) -> PIDevice:
        """Select the correct VISA device driver based on the ``*IDN?`` response.

        Be sure to handle removing the device from the config on a SystemError.

        Args:
            visa_resource: The VISA resource object.
            device_config: The device configuration object.
            device_drivers: The mapping of model series strings to device drivers.

        Returns:
            The new device driver.

        Raises:
            SystemError: Indicating the correct driver was unable to be determined.
        """
        idn_response = self.__clear_visa_output_buffer_and_get_idn(visa_resource)
        model_series = ""
        try:
            model_series = get_model_series(idn_response.split(",")[1])
            device_driver = cast(Type[PIDevice], device_drivers[model_series])
            new_device = device_driver(device_config, self.__verbose, visa_resource)
        except (KeyError, IndexError) as error:
            visa_resource.close()
            message = "Unable to determine which device driver to use."
            if model_series:
                message += f' The parsed model series is "{model_series}".'
            message += f" *IDN? returned {idn_response!r}"
            raise SystemError(message) from error

        return new_device

    def __set_options(self, verbose: bool) -> None:
        """Set the options for the DeviceManager.

        Args:
            verbose: A boolean indicating if verbose output should be printed.
        """
        self.__verbose = self.__config.options.verbose_mode = verbose or bool(
            self.__config.options.verbose_mode
        )
        self.__teardown_cleanup_enabled = bool(self.__config.options.teardown_cleanup)
        self.__setup_cleanup_enabled = bool(self.__config.options.setup_cleanup)
        # Configure the VISA library
        if self.__config.options.standalone:
            self.__visa_library = PYVISA_PY_BACKEND
        elif self.__visa_library is NotImplemented:
            # Check for an environment variable, primarily used for enabling
            # validation of example code on simulated instruments.
            if visa_library := os.getenv("TM_DEVICES_VISA_LIBRARY"):
                self.__visa_library = visa_library
            else:
                self.__visa_library = ""
        self.verbose_visa = bool(self.__config.options.verbose_visa)


def print_available_visa_devices() -> None:  # pragma: no cover
    """Print all available VISA devices to the console."""
    with contextlib.redirect_stdout(None), contextlib.redirect_stderr(None), DeviceManager() as dm:
        available_devices = dm.get_available_devices()
    print(json.dumps(available_devices["local"], indent=2))



"""MSO4 device driver module."""
from tm_devices.commands import MSO4Mixin
from tm_devices.drivers.pi.scopes.tekscope.tekscope import TekScope


class MSO4(MSO4Mixin, TekScope):
    """MSO4 device driver."""



"""SMU Model 2657A device driver module."""
from tm_devices.commands import SMU2657AMixin
from tm_devices.drivers.pi.source_measure_units.smu2600.smu2650a import SMU2650A


class SMU2657A(SMU2657AMixin, SMU2650A):
    """SMU2657A device driver."""



"""The status_and_error commands module.

These commands are used in the following models:
DPO2K, DPO2KB, DPO4K, DPO4KB, MDO3, MDO3K, MDO4K, MDO4KB, MDO4KC, MSO2K, MSO2KB, MSO4K, MSO4KB

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:

::

    - *PSC {OFF|ON|NR1>}
    - *PSC?
"""
from typing import Optional, TYPE_CHECKING

from .._helpers import SCPICmdRead, SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class Psc(SCPICmdWrite, SCPICmdRead):
    """The ``*PSC`` command.

    **Description:**
        - This command specifies the power-on status flag that controls the automatic power-on
          handling of the DESER, SRER, and ESER registers. When ``*PSC`` is true, the DESER register
          is set to 255 and the SRER and ESER registers are set to 0 at power-on. When ``*PSC`` is
          false, the current values in the DESER, SRER, and ESER registers are preserved in
          nonvolatile memory when power is shut off and are restored at power-on.

    **Usage:**
        - Using the ``.query()`` method will send the ``*PSC?`` query.
        - Using the ``.verify(value)`` method will send the ``*PSC?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``*PSC value`` command.

    **SCPI Syntax:**

    ::

        - *PSC {OFF|ON|NR1>}
        - *PSC?

    **Info:**
        - ``OFF`` sets the power-on status clear flag to false.
        - ``ON`` sets the power-on status clear flag to true.
        - ``<NR1>`` = 0 sets the power-on status clear flag to false. This disables the power-on
          clear allowing the oscilloscope to possibly assert SRQ after power-on; any other value
          sets the power-on status clear flag to true, enabling the power-on status clear preventing
          any SRQ assertion after power on.
    """

    def __init__(self, device: Optional["PIDevice"] = None, cmd_syntax: str = "*PSC") -> None:
        super().__init__(device, cmd_syntax)



"""The password commands module.

These commands are used in the following models:
DPO2K, DPO2KB, DPO4K, DPO4KB, DPO5KB, DPO70KC, DPO70KD, DPO70KDX, DPO70KSX, DPO7KC, DSA70KC,
DSA70KD, LPD6, MDO3, MDO3K, MDO4K, MDO4KB, MDO4KC, MSO2, MSO2K, MSO2KB, MSO4, MSO4K, MSO4KB, MSO5,
MSO5B, MSO5KB, MSO5LP, MSO6, MSO6B, MSO70KC, MSO70KDX

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:

::

    - PASSWord <QString>
"""
from typing import Optional, TYPE_CHECKING

from .._helpers import SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class Password(SCPICmdWrite):
    """The ``PASSWord`` command.

    **Description:**
        - This command (no query form) enables the ``*PUD`` and NEWpass set commands. Sending
          ``PASSWord`` without any arguments disables these same commands. Once the password is
          successfully entered, the ``*PUD`` and NEWpass commands are enabled until the instrument
          is powered off, or until the FACtory command, the ``PASSWord`` command with no arguments,
          or the ``*RST`` command is issued. To change the password, you must first enter the valid
          password with the ``PASSWord`` command and then change to your new password with the
          NEWpass command. Remember that the password is case sensitive.

    **Usage:**
        - Using the ``.write(value)`` method will send the ``PASSWord value`` command.

    **SCPI Syntax:**

    ::

        - PASSWord <QString>

    **Info:**
        - ``<QString>`` is the password, which can contain up to 10 characters. The factory default
          password is 'XYZZY' and is always valid.
    """

    _WRAP_ARG_WITH_QUOTES = True

    def __init__(self, device: Optional["PIDevice"] = None, cmd_syntax: str = "PASSWord") -> None:
        super().__init__(device, cmd_syntax)



"""AFG3K device driver module."""
from tm_devices.commands import AFG3KMixin
from tm_devices.drivers.pi.signal_sources.afgs.afg import AFG, AFGSourceDeviceConstants


class AFG3K(AFG3KMixin, AFG):
    """AFG3K device driver."""

    _DEVICE_CONSTANTS = AFGSourceDeviceConstants(
        memory_page_size=2,
        memory_max_record_length=128 * 1024,
        memory_min_record_length=2,
    )



"""The source commands module.

These commands are used in the following models:
AFG3K, AFG3KB, AFG3KC

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:

::

    - SOURce:ROSCillator:SOURce {INTernal|EXTernal}
    - SOURce:ROSCillator:SOURce?
"""
from typing import Optional, TYPE_CHECKING

from .._helpers import SCPICmdRead, SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class SourceRoscillatorSource(SCPICmdWrite, SCPICmdRead):
    """The ``SOURce:ROSCillator:SOURce`` command.

    **Description:**
        - This command sets the reference clock to either internal or external.

    **Usage:**
        - Using the ``.query()`` method will send the ``SOURce:ROSCillator:SOURce?`` query.
        - Using the ``.verify(value)`` method will send the ``SOURce:ROSCillator:SOURce?`` query and
          raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SOURce:ROSCillator:SOURce value``
          command.

    **SCPI Syntax:**

    ::

        - SOURce:ROSCillator:SOURce {INTernal|EXTernal}
        - SOURce:ROSCillator:SOURce?
    """


class SourceRoscillator(SCPICmdRead):
    """The ``SOURce:ROSCillator`` command tree.

    **Usage:**
        - Using the ``.query()`` method will send the ``SOURce:ROSCillator?`` query.
        - Using the ``.verify(value)`` method will send the ``SOURce:ROSCillator?`` query and raise
          an AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.source``: The ``SOURce:ROSCillator:SOURce`` command.
    """

    def __init__(self, device: Optional["PIDevice"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._source = SourceRoscillatorSource(device, f"{self._cmd_syntax}:SOURce")

    @property
    def source(self) -> SourceRoscillatorSource:
        """Return the ``SOURce:ROSCillator:SOURce`` command.

        **Description:**
            - This command sets the reference clock to either internal or external.

        **Usage:**
            - Using the ``.query()`` method will send the ``SOURce:ROSCillator:SOURce?`` query.
            - Using the ``.verify(value)`` method will send the ``SOURce:ROSCillator:SOURce?`` query
              and raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SOURce:ROSCillator:SOURce value``
              command.

        **SCPI Syntax:**

        ::

            - SOURce:ROSCillator:SOURce {INTernal|EXTernal}
            - SOURce:ROSCillator:SOURce?
        """
        return self._source


class Source(SCPICmdRead):
    """The ``SOURce`` command tree.

    **Usage:**
        - Using the ``.query()`` method will send the ``SOURce?`` query.
        - Using the ``.verify(value)`` method will send the ``SOURce?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.roscillator``: The ``SOURce:ROSCillator`` command tree.
    """

    def __init__(self, device: Optional["PIDevice"] = None, cmd_syntax: str = "SOURce") -> None:
        super().__init__(device, cmd_syntax)
        self._roscillator = SourceRoscillator(device, f"{self._cmd_syntax}:ROSCillator")

    @property
    def roscillator(self) -> SourceRoscillator:
        """Return the ``SOURce:ROSCillator`` command tree.

        **Usage:**
            - Using the ``.query()`` method will send the ``SOURce:ROSCillator?`` query.
            - Using the ``.verify(value)`` method will send the ``SOURce:ROSCillator?`` query and
              raise an AssertionError if the returned value does not match ``value``.

        Sub-properties:
            - ``.source``: The ``SOURce:ROSCillator:SOURce`` command.
        """
        return self._roscillator



"""The socketserver commands module.

These commands are used in the following models:
LPD6, MSO2, MSO4, MSO5, MSO5B, MSO5LP, MSO6, MSO6B

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:

::

    - SOCKETServer:ENAble {0|1|OFF|ON}
    - SOCKETServer:ENAble?
    - SOCKETServer:PORT <NR1>
    - SOCKETServer:PORT?
    - SOCKETServer:PROTOCol {TERMinal|NONe}
    - SOCKETServer:PROTOCol?
"""
from typing import Optional, TYPE_CHECKING

from .._helpers import SCPICmdRead, SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class SocketserverProtocol(SCPICmdWrite, SCPICmdRead):
    """The ``SOCKETServer:PROTOCol`` command.

    **Description:**
        - This command sets or queries the protocol for the socket server.

    **Usage:**
        - Using the ``.query()`` method will send the ``SOCKETServer:PROTOCol?`` query.
        - Using the ``.verify(value)`` method will send the ``SOCKETServer:PROTOCol?`` query and
          raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SOCKETServer:PROTOCol value`` command.

    **SCPI Syntax:**

    ::

        - SOCKETServer:PROTOCol {TERMinal|NONe}
        - SOCKETServer:PROTOCol?

    **Info:**
        - ``TERMinal`` specifies terminal protocol for the socket server. When set to TERMinal, a
          session startup message is sent to the socket and a command prompt is provided.
        - ``NONe`` disables the terminal features, allowing the server to be used for raw socket
          transactions, such as with a VISA socket server. The default setting is NONe.
    """


class SocketserverPort(SCPICmdWrite, SCPICmdRead):
    """The ``SOCKETServer:PORT`` command.

    **Description:**
        - This command sets the TCPIP port for the socket server connection.

    **Usage:**
        - Using the ``.query()`` method will send the ``SOCKETServer:PORT?`` query.
        - Using the ``.verify(value)`` method will send the ``SOCKETServer:PORT?`` query and raise
          an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SOCKETServer:PORT value`` command.

    **SCPI Syntax:**

    ::

        - SOCKETServer:PORT <NR1>
        - SOCKETServer:PORT?

    **Info:**
        - ``<NR1>`` is the TCPIP port for the socket server connection.
    """


class SocketserverEnable(SCPICmdWrite, SCPICmdRead):
    """The ``SOCKETServer:ENAble`` command.

    **Description:**
        - This command enables or disables the socket server which supports a telnet or other TCPIP
          socket connection to send commands and queries to the instrument. The default state is
          enabled.

    **Usage:**
        - Using the ``.query()`` method will send the ``SOCKETServer:ENAble?`` query.
        - Using the ``.verify(value)`` method will send the ``SOCKETServer:ENAble?`` query and raise
          an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``SOCKETServer:ENAble value`` command.

    **SCPI Syntax:**

    ::

        - SOCKETServer:ENAble {0|1|OFF|ON}
        - SOCKETServer:ENAble?

    **Info:**
        - ``1`` enables the socket server. If the state is 0 (disabled) and this command is sent to
          enable the socket server when the port is in use by another service, then the error event
          code 221 (Settings conflict) is posted to the event queue and the socket server remains
          disabled. In this case, select a different port number and attempt to enable the socket
          server again.
        - ``0`` disables the socket server.
        - ``ON`` enables the socket server.
        - ``OFF`` disables the socket server.
    """


class Socketserver(SCPICmdRead):
    """The ``SOCKETServer`` command tree.

    **Usage:**
        - Using the ``.query()`` method will send the ``SOCKETServer?`` query.
        - Using the ``.verify(value)`` method will send the ``SOCKETServer?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.enable``: The ``SOCKETServer:ENAble`` command.
        - ``.port``: The ``SOCKETServer:PORT`` command.
        - ``.protocol``: The ``SOCKETServer:PROTOCol`` command.
    """

    def __init__(
        self, device: Optional["PIDevice"] = None, cmd_syntax: str = "SOCKETServer"
    ) -> None:
        super().__init__(device, cmd_syntax)
        self._enable = SocketserverEnable(device, f"{self._cmd_syntax}:ENAble")
        self._port = SocketserverPort(device, f"{self._cmd_syntax}:PORT")
        self._protocol = SocketserverProtocol(device, f"{self._cmd_syntax}:PROTOCol")

    @property
    def enable(self) -> SocketserverEnable:
        """Return the ``SOCKETServer:ENAble`` command.

        **Description:**
            - This command enables or disables the socket server which supports a telnet or other
              TCPIP socket connection to send commands and queries to the instrument. The default
              state is enabled.

        **Usage:**
            - Using the ``.query()`` method will send the ``SOCKETServer:ENAble?`` query.
            - Using the ``.verify(value)`` method will send the ``SOCKETServer:ENAble?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SOCKETServer:ENAble value``
              command.

        **SCPI Syntax:**

        ::

            - SOCKETServer:ENAble {0|1|OFF|ON}
            - SOCKETServer:ENAble?

        **Info:**
            - ``1`` enables the socket server. If the state is 0 (disabled) and this command is sent
              to enable the socket server when the port is in use by another service, then the error
              event code 221 (Settings conflict) is posted to the event queue and the socket server
              remains disabled. In this case, select a different port number and attempt to enable
              the socket server again.
            - ``0`` disables the socket server.
            - ``ON`` enables the socket server.
            - ``OFF`` disables the socket server.
        """
        return self._enable

    @property
    def port(self) -> SocketserverPort:
        """Return the ``SOCKETServer:PORT`` command.

        **Description:**
            - This command sets the TCPIP port for the socket server connection.

        **Usage:**
            - Using the ``.query()`` method will send the ``SOCKETServer:PORT?`` query.
            - Using the ``.verify(value)`` method will send the ``SOCKETServer:PORT?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SOCKETServer:PORT value`` command.

        **SCPI Syntax:**

        ::

            - SOCKETServer:PORT <NR1>
            - SOCKETServer:PORT?

        **Info:**
            - ``<NR1>`` is the TCPIP port for the socket server connection.
        """
        return self._port

    @property
    def protocol(self) -> SocketserverProtocol:
        """Return the ``SOCKETServer:PROTOCol`` command.

        **Description:**
            - This command sets or queries the protocol for the socket server.

        **Usage:**
            - Using the ``.query()`` method will send the ``SOCKETServer:PROTOCol?`` query.
            - Using the ``.verify(value)`` method will send the ``SOCKETServer:PROTOCol?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``SOCKETServer:PROTOCol value``
              command.

        **SCPI Syntax:**

        ::

            - SOCKETServer:PROTOCol {TERMinal|NONe}
            - SOCKETServer:PROTOCol?

        **Info:**
            - ``TERMinal`` specifies terminal protocol for the socket server. When set to TERMinal,
              a session startup message is sent to the socket and a command prompt is provided.
            - ``NONe`` disables the terminal features, allowing the server to be used for raw socket
              transactions, such as with a VISA socket server. The default setting is NONe.
        """
        return self._protocol



"""The acquire commands module.

These commands are used in the following models:
MSO2

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:

::

    - ACQuire:MAXSamplerate?
    - ACQuire:MODe {SAMple|PEAKdetect|HIRes|AVErage|ENVelope}
    - ACQuire:MODe?
    - ACQuire:NUMACq?
    - ACQuire:NUMAVg <NR1>
    - ACQuire:NUMAVg?
    - ACQuire:SEQuence:CURrent?
    - ACQuire:SEQuence:MODe NUMACQs
    - ACQuire:SEQuence:NUMSEQuence <NR1>
    - ACQuire:SEQuence:NUMSEQuence?
    - ACQuire:STATE {<NR1>|OFF|ON|RUN|STOP}
    - ACQuire:STATE?
    - ACQuire:STOPAfter {RUNSTop|SEQuence}
    - ACQuire:STOPAfter?
    - ACQuire?
"""
from typing import Optional, TYPE_CHECKING

from .._helpers import SCPICmdRead, SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class AcquireStopafter(SCPICmdWrite, SCPICmdRead):
    """The ``ACQuire:STOPAfter`` command.

    **Description:**
        - This command sets or queries whether the instrument continually acquires acquisitions or
          acquires a single sequence. Pressing SINGLE on the front panel button is equivalent to
          sending these commands: ``ACQUIRE:STOPAFTER SEQUENCE`` and ``ACQUIRE:STATE 1``.

    **Usage:**
        - Using the ``.query()`` method will send the ``ACQuire:STOPAfter?`` query.
        - Using the ``.verify(value)`` method will send the ``ACQuire:STOPAfter?`` query and raise
          an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``ACQuire:STOPAfter value`` command.

    **SCPI Syntax:**

    ::

        - ACQuire:STOPAfter {RUNSTop|SEQuence}
        - ACQuire:STOPAfter?

    **Info:**
        - ``RUNSTop`` specifies that the instrument will continually acquire data, if
          ``ACQuire:STATE`` is turned on.
        - ``SEQuence`` specifies that the next acquisition will be a single-sequence acquisition.
    """


class AcquireState(SCPICmdWrite, SCPICmdRead):
    """The ``ACQuire:STATE`` command.

    **Description:**
        - This command starts or stops acquisitions. When state is set to ON or RUN, a new
          acquisition will be started. If the last acquisition was a single acquisition sequence, a
          new single sequence acquisition will be started. If the last acquisition was continuous, a
          new continuous acquisition will be started. If RUN is issued in the middle of completing a
          single sequence acquisition (for example, averaging or enveloping), the acquisition
          sequence is restarted, and any accumulated data is discarded. Also, the instrument resets
          the number of acquisitions. If the RUN argument is issued while in continuous mode, a
          reset occurs and acquired data continues to acquire. If ``acquire:stopafter`` is SEQUENCE,
          this command leaves the instrument in single sequence, unlike the run/stop button which
          takes the instrument out of single sequence.

    **Usage:**
        - Using the ``.query()`` method will send the ``ACQuire:STATE?`` query.
        - Using the ``.verify(value)`` method will send the ``ACQuire:STATE?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``ACQuire:STATE value`` command.

    **SCPI Syntax:**

    ::

        - ACQuire:STATE {<NR1>|OFF|ON|RUN|STOP}
        - ACQuire:STATE?

    **Info:**
        - ``<NR1>`` = 0 stops acquisitions; any other value starts acquisitions.
        - ``OFF`` stops acquisitions.
        - ``ON`` starts acquisitions.
        - ``RUN`` starts acquisitions.
        - ``STOP`` stops acquisitions.
    """


class AcquireSequenceNumsequence(SCPICmdWrite, SCPICmdRead):
    """The ``ACQuire:SEQuence:NUMSEQuence`` command.

    **Description:**
        - In single sequence acquisition mode, specify the number of acquisitions or measurements
          that comprise the sequence. The default is 1.

    **Usage:**
        - Using the ``.query()`` method will send the ``ACQuire:SEQuence:NUMSEQuence?`` query.
        - Using the ``.verify(value)`` method will send the ``ACQuire:SEQuence:NUMSEQuence?`` query
          and raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``ACQuire:SEQuence:NUMSEQuence value``
          command.

    **SCPI Syntax:**

    ::

        - ACQuire:SEQuence:NUMSEQuence <NR1>
        - ACQuire:SEQuence:NUMSEQuence?

    **Info:**
        - ``<NR1>`` is the number of acquisitions or measurements that comprise the sequence.
    """


class AcquireSequenceMode(SCPICmdWrite):
    """The ``ACQuire:SEQuence:MODe`` command.

    **Description:**
        - In single sequence acquisition, the single sequence stop after count is based on the
          number of acquisitions.

    **Usage:**
        - Using the ``.write(value)`` method will send the ``ACQuire:SEQuence:MODe value`` command.

    **SCPI Syntax:**

    ::

        - ACQuire:SEQuence:MODe NUMACQs

    **Info:**
        - ``NUMACQs`` is the number of acquisitions.
    """


class AcquireSequenceCurrent(SCPICmdRead):
    """The ``ACQuire:SEQuence:CURrent`` command.

    **Description:**
        - In single sequence acquisition mode, this query returns the number of acquisitions or
          measurements in the sequence completed so far.

    **Usage:**
        - Using the ``.query()`` method will send the ``ACQuire:SEQuence:CURrent?`` query.
        - Using the ``.verify(value)`` method will send the ``ACQuire:SEQuence:CURrent?`` query and
          raise an AssertionError if the returned value does not match ``value``.

    **SCPI Syntax:**

    ::

        - ACQuire:SEQuence:CURrent?
    """


class AcquireSequence(SCPICmdRead):
    """The ``ACQuire:SEQuence`` command tree.

    **Usage:**
        - Using the ``.query()`` method will send the ``ACQuire:SEQuence?`` query.
        - Using the ``.verify(value)`` method will send the ``ACQuire:SEQuence?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.current``: The ``ACQuire:SEQuence:CURrent`` command.
        - ``.mode``: The ``ACQuire:SEQuence:MODe`` command.
        - ``.numsequence``: The ``ACQuire:SEQuence:NUMSEQuence`` command.
    """

    def __init__(self, device: Optional["PIDevice"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._current = AcquireSequenceCurrent(device, f"{self._cmd_syntax}:CURrent")
        self._mode = AcquireSequenceMode(device, f"{self._cmd_syntax}:MODe")
        self._numsequence = AcquireSequenceNumsequence(device, f"{self._cmd_syntax}:NUMSEQuence")

    @property
    def current(self) -> AcquireSequenceCurrent:
        """Return the ``ACQuire:SEQuence:CURrent`` command.

        **Description:**
            - In single sequence acquisition mode, this query returns the number of acquisitions or
              measurements in the sequence completed so far.

        **Usage:**
            - Using the ``.query()`` method will send the ``ACQuire:SEQuence:CURrent?`` query.
            - Using the ``.verify(value)`` method will send the ``ACQuire:SEQuence:CURrent?`` query
              and raise an AssertionError if the returned value does not match ``value``.

        **SCPI Syntax:**

        ::

            - ACQuire:SEQuence:CURrent?
        """
        return self._current

    @property
    def mode(self) -> AcquireSequenceMode:
        """Return the ``ACQuire:SEQuence:MODe`` command.

        **Description:**
            - In single sequence acquisition, the single sequence stop after count is based on the
              number of acquisitions.

        **Usage:**
            - Using the ``.write(value)`` method will send the ``ACQuire:SEQuence:MODe value``
              command.

        **SCPI Syntax:**

        ::

            - ACQuire:SEQuence:MODe NUMACQs

        **Info:**
            - ``NUMACQs`` is the number of acquisitions.
        """
        return self._mode

    @property
    def numsequence(self) -> AcquireSequenceNumsequence:
        """Return the ``ACQuire:SEQuence:NUMSEQuence`` command.

        **Description:**
            - In single sequence acquisition mode, specify the number of acquisitions or
              measurements that comprise the sequence. The default is 1.

        **Usage:**
            - Using the ``.query()`` method will send the ``ACQuire:SEQuence:NUMSEQuence?`` query.
            - Using the ``.verify(value)`` method will send the ``ACQuire:SEQuence:NUMSEQuence?``
              query and raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the
              ``ACQuire:SEQuence:NUMSEQuence value`` command.

        **SCPI Syntax:**

        ::

            - ACQuire:SEQuence:NUMSEQuence <NR1>
            - ACQuire:SEQuence:NUMSEQuence?

        **Info:**
            - ``<NR1>`` is the number of acquisitions or measurements that comprise the sequence.
        """
        return self._numsequence


class AcquireNumavg(SCPICmdWrite, SCPICmdRead):
    """The ``ACQuire:NUMAVg`` command.

    **Description:**
        - This command sets or queries the number of waveform acquisitions that make up an averaged
          waveform. Ranges from 2 to 10240.

    **Usage:**
        - Using the ``.query()`` method will send the ``ACQuire:NUMAVg?`` query.
        - Using the ``.verify(value)`` method will send the ``ACQuire:NUMAVg?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``ACQuire:NUMAVg value`` command.

    **SCPI Syntax:**

    ::

        - ACQuire:NUMAVg <NR1>
        - ACQuire:NUMAVg?

    **Info:**
        - ``<NR1>`` is the number of waveform acquisitions to average.
    """


class AcquireNumacq(SCPICmdRead):
    """The ``ACQuire:NUMACq`` command.

    **Description:**
        - This query-only command returns the number of waveform acquisitions that have occurred
          since the last time acquisitions were stopped.

    **Usage:**
        - Using the ``.query()`` method will send the ``ACQuire:NUMACq?`` query.
        - Using the ``.verify(value)`` method will send the ``ACQuire:NUMACq?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    **SCPI Syntax:**

    ::

        - ACQuire:NUMACq?
    """


class AcquireMode(SCPICmdWrite, SCPICmdRead):
    """The ``ACQuire:MODe`` command.

    **Description:**
        - This command sets or queries the selected acquisition mode of the instrument.

    **Usage:**
        - Using the ``.query()`` method will send the ``ACQuire:MODe?`` query.
        - Using the ``.verify(value)`` method will send the ``ACQuire:MODe?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``ACQuire:MODe value`` command.

    **SCPI Syntax:**

    ::

        - ACQuire:MODe {SAMple|PEAKdetect|HIRes|AVErage|ENVelope}
        - ACQuire:MODe?

    **Info:**
        - ``SAMple`` specifies that the displayed data point value is the first sampled value that
          is taken during the acquisition interval. The CURVE query, depending on sample rate, will
          result in either 8 bit or 16 bit data. In case of 8 bit data, the precision is also 8 bit.
          However, in case of 16 bit data, 12 bit precision data is zero padded in lower 4 bits.
          SAMple is the default mode.
        - ``PEAKdetect`` specifies the display of high-low range of the samples taken from a single
          waveform acquisition. The high-low range is displayed as a vertical column that extends
          from the highest to the lowest value sampled during the acquisition interval. PEAKdetect
          mode can reveal the presence of aliasing or narrow spikes.
        - ``HIRes`` specifies Hi Res mode where the displayed data point value is the average of all
          the samples taken during the acquisition interval. This is a form of averaging, where the
          average comes from a single waveform acquisition. The number of samples taken during the
          acquisition interval determines the number of data values that compose the average.
        - ``AVErage`` specifies averaging mode, in which the resulting waveform shows an average of
          SAMple data points from several separate waveform acquisitions. The instrument processes
          the number of waveforms you specify into the acquired waveform, creating a running
          exponential average of the input signal. The number of waveform acquisitions that go into
          making up the average waveform is set or queried using the ``ACQUIRE:NUMAVG`` command.
        - ``ENVelope`` specifies envelope mode, where the resulting waveform displays the range of
          PEAKdetect from continued waveform acquisitions.
    """


class AcquireMaxsamplerate(SCPICmdRead):
    """The ``ACQuire:MAXSamplerate`` command.

    **Description:**
        - This query returns the maximum real-time sample rate.

    **Usage:**
        - Using the ``.query()`` method will send the ``ACQuire:MAXSamplerate?`` query.
        - Using the ``.verify(value)`` method will send the ``ACQuire:MAXSamplerate?`` query and
          raise an AssertionError if the returned value does not match ``value``.

    **SCPI Syntax:**

    ::

        - ACQuire:MAXSamplerate?
    """


class Acquire(SCPICmdRead):
    """The ``ACQuire`` command.

    **Description:**
        - Queries the current acquisition state.

    **Usage:**
        - Using the ``.query()`` method will send the ``ACQuire?`` query.
        - Using the ``.verify(value)`` method will send the ``ACQuire?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    **SCPI Syntax:**

    ::

        - ACQuire?

    Properties:
        - ``.maxsamplerate``: The ``ACQuire:MAXSamplerate`` command.
        - ``.mode``: The ``ACQuire:MODe`` command.
        - ``.numacq``: The ``ACQuire:NUMACq`` command.
        - ``.numavg``: The ``ACQuire:NUMAVg`` command.
        - ``.sequence``: The ``ACQuire:SEQuence`` command tree.
        - ``.state``: The ``ACQuire:STATE`` command.
        - ``.stopafter``: The ``ACQuire:STOPAfter`` command.
    """

    def __init__(self, device: Optional["PIDevice"] = None, cmd_syntax: str = "ACQuire") -> None:
        super().__init__(device, cmd_syntax)
        self._maxsamplerate = AcquireMaxsamplerate(device, f"{self._cmd_syntax}:MAXSamplerate")
        self._mode = AcquireMode(device, f"{self._cmd_syntax}:MODe")
        self._numacq = AcquireNumacq(device, f"{self._cmd_syntax}:NUMACq")
        self._numavg = AcquireNumavg(device, f"{self._cmd_syntax}:NUMAVg")
        self._sequence = AcquireSequence(device, f"{self._cmd_syntax}:SEQuence")
        self._state = AcquireState(device, f"{self._cmd_syntax}:STATE")
        self._stopafter = AcquireStopafter(device, f"{self._cmd_syntax}:STOPAfter")

    @property
    def maxsamplerate(self) -> AcquireMaxsamplerate:
        """Return the ``ACQuire:MAXSamplerate`` command.

        **Description:**
            - This query returns the maximum real-time sample rate.

        **Usage:**
            - Using the ``.query()`` method will send the ``ACQuire:MAXSamplerate?`` query.
            - Using the ``.verify(value)`` method will send the ``ACQuire:MAXSamplerate?`` query and
              raise an AssertionError if the returned value does not match ``value``.

        **SCPI Syntax:**

        ::

            - ACQuire:MAXSamplerate?
        """
        return self._maxsamplerate

    @property
    def mode(self) -> AcquireMode:
        """Return the ``ACQuire:MODe`` command.

        **Description:**
            - This command sets or queries the selected acquisition mode of the instrument.

        **Usage:**
            - Using the ``.query()`` method will send the ``ACQuire:MODe?`` query.
            - Using the ``.verify(value)`` method will send the ``ACQuire:MODe?`` query and raise an
              AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``ACQuire:MODe value`` command.

        **SCPI Syntax:**

        ::

            - ACQuire:MODe {SAMple|PEAKdetect|HIRes|AVErage|ENVelope}
            - ACQuire:MODe?

        **Info:**
            - ``SAMple`` specifies that the displayed data point value is the first sampled value
              that is taken during the acquisition interval. The CURVE query, depending on sample
              rate, will result in either 8 bit or 16 bit data. In case of 8 bit data, the precision
              is also 8 bit. However, in case of 16 bit data, 12 bit precision data is zero padded
              in lower 4 bits. SAMple is the default mode.
            - ``PEAKdetect`` specifies the display of high-low range of the samples taken from a
              single waveform acquisition. The high-low range is displayed as a vertical column that
              extends from the highest to the lowest value sampled during the acquisition interval.
              PEAKdetect mode can reveal the presence of aliasing or narrow spikes.
            - ``HIRes`` specifies Hi Res mode where the displayed data point value is the average of
              all the samples taken during the acquisition interval. This is a form of averaging,
              where the average comes from a single waveform acquisition. The number of samples
              taken during the acquisition interval determines the number of data values that
              compose the average.
            - ``AVErage`` specifies averaging mode, in which the resulting waveform shows an average
              of SAMple data points from several separate waveform acquisitions. The instrument
              processes the number of waveforms you specify into the acquired waveform, creating a
              running exponential average of the input signal. The number of waveform acquisitions
              that go into making up the average waveform is set or queried using the
              ``ACQUIRE:NUMAVG`` command.
            - ``ENVelope`` specifies envelope mode, where the resulting waveform displays the range
              of PEAKdetect from continued waveform acquisitions.
        """
        return self._mode

    @property
    def numacq(self) -> AcquireNumacq:
        """Return the ``ACQuire:NUMACq`` command.

        **Description:**
            - This query-only command returns the number of waveform acquisitions that have occurred
              since the last time acquisitions were stopped.

        **Usage:**
            - Using the ``.query()`` method will send the ``ACQuire:NUMACq?`` query.
            - Using the ``.verify(value)`` method will send the ``ACQuire:NUMACq?`` query and raise
              an AssertionError if the returned value does not match ``value``.

        **SCPI Syntax:**

        ::

            - ACQuire:NUMACq?
        """
        return self._numacq

    @property
    def numavg(self) -> AcquireNumavg:
        """Return the ``ACQuire:NUMAVg`` command.

        **Description:**
            - This command sets or queries the number of waveform acquisitions that make up an
              averaged waveform. Ranges from 2 to 10240.

        **Usage:**
            - Using the ``.query()`` method will send the ``ACQuire:NUMAVg?`` query.
            - Using the ``.verify(value)`` method will send the ``ACQuire:NUMAVg?`` query and raise
              an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``ACQuire:NUMAVg value`` command.

        **SCPI Syntax:**

        ::

            - ACQuire:NUMAVg <NR1>
            - ACQuire:NUMAVg?

        **Info:**
            - ``<NR1>`` is the number of waveform acquisitions to average.
        """
        return self._numavg

    @property
    def sequence(self) -> AcquireSequence:
        """Return the ``ACQuire:SEQuence`` command tree.

        **Usage:**
            - Using the ``.query()`` method will send the ``ACQuire:SEQuence?`` query.
            - Using the ``.verify(value)`` method will send the ``ACQuire:SEQuence?`` query and
              raise an AssertionError if the returned value does not match ``value``.

        Sub-properties:
            - ``.current``: The ``ACQuire:SEQuence:CURrent`` command.
            - ``.mode``: The ``ACQuire:SEQuence:MODe`` command.
            - ``.numsequence``: The ``ACQuire:SEQuence:NUMSEQuence`` command.
        """
        return self._sequence

    @property
    def state(self) -> AcquireState:
        """Return the ``ACQuire:STATE`` command.

        **Description:**
            - This command starts or stops acquisitions. When state is set to ON or RUN, a new
              acquisition will be started. If the last acquisition was a single acquisition
              sequence, a new single sequence acquisition will be started. If the last acquisition
              was continuous, a new continuous acquisition will be started. If RUN is issued in the
              middle of completing a single sequence acquisition (for example, averaging or
              enveloping), the acquisition sequence is restarted, and any accumulated data is
              discarded. Also, the instrument resets the number of acquisitions. If the RUN argument
              is issued while in continuous mode, a reset occurs and acquired data continues to
              acquire. If ``acquire:stopafter`` is SEQUENCE, this command leaves the instrument in
              single sequence, unlike the run/stop button which takes the instrument out of single
              sequence.

        **Usage:**
            - Using the ``.query()`` method will send the ``ACQuire:STATE?`` query.
            - Using the ``.verify(value)`` method will send the ``ACQuire:STATE?`` query and raise
              an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``ACQuire:STATE value`` command.

        **SCPI Syntax:**

        ::

            - ACQuire:STATE {<NR1>|OFF|ON|RUN|STOP}
            - ACQuire:STATE?

        **Info:**
            - ``<NR1>`` = 0 stops acquisitions; any other value starts acquisitions.
            - ``OFF`` stops acquisitions.
            - ``ON`` starts acquisitions.
            - ``RUN`` starts acquisitions.
            - ``STOP`` stops acquisitions.
        """
        return self._state

    @property
    def stopafter(self) -> AcquireStopafter:
        """Return the ``ACQuire:STOPAfter`` command.

        **Description:**
            - This command sets or queries whether the instrument continually acquires acquisitions
              or acquires a single sequence. Pressing SINGLE on the front panel button is equivalent
              to sending these commands: ``ACQUIRE:STOPAFTER SEQUENCE`` and ``ACQUIRE:STATE 1``.

        **Usage:**
            - Using the ``.query()`` method will send the ``ACQuire:STOPAfter?`` query.
            - Using the ``.verify(value)`` method will send the ``ACQuire:STOPAfter?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``ACQuire:STOPAfter value`` command.

        **SCPI Syntax:**

        ::

            - ACQuire:STOPAfter {RUNSTop|SEQuence}
            - ACQuire:STOPAfter?

        **Info:**
            - ``RUNSTop`` specifies that the instrument will continually acquire data, if
              ``ACQuire:STATE`` is turned on.
            - ``SEQuence`` specifies that the next acquisition will be a single-sequence
              acquisition.
        """
        return self._stopafter



"""The trig commands module.

These commands are used in the following models:
DPO5KB, DPO70KC, DPO70KD, DPO70KDX, DPO70KSX, DPO7KC, DSA70KC, DSA70KD, MSO5KB, MSO70KC, MSO70KDX

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:

::

    - TRIG:A:PLOCK:STANDARD {<STANDARD>}
    - TRIG:A:PLOCK:STANDARD?
"""
from typing import Optional, TYPE_CHECKING

from .._helpers import SCPICmdRead, SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class TrigAPlockStandard(SCPICmdWrite, SCPICmdRead):
    """The ``TRIG:A:PLOCK:STANDARD`` command.

    **Description:**
        - This command sets or queries the communication standard used by the pattern lock trigger.

    **Usage:**
        - Using the ``.query()`` method will send the ``TRIG:A:PLOCK:STANDARD?`` query.
        - Using the ``.verify(value)`` method will send the ``TRIG:A:PLOCK:STANDARD?`` query and
          raise an AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``TRIG:A:PLOCK:STANDARD value`` command.

    **SCPI Syntax:**

    ::

        - TRIG:A:PLOCK:STANDARD {<STANDARD>}
        - TRIG:A:PLOCK:STANDARD?

    **Info:**
        - ``STANDARD`` communication standard used by the pattern lock trigger. See
          ``TRIGGER:A:COMMUNICATION:STANDARD`` for a list of standards.
    """


class TrigAPlock(SCPICmdRead):
    """The ``TRIG:A:PLOCK`` command tree.

    **Usage:**
        - Using the ``.query()`` method will send the ``TRIG:A:PLOCK?`` query.
        - Using the ``.verify(value)`` method will send the ``TRIG:A:PLOCK?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.standard``: The ``TRIG:A:PLOCK:STANDARD`` command.
    """

    def __init__(self, device: Optional["PIDevice"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._standard = TrigAPlockStandard(device, f"{self._cmd_syntax}:STANDARD")

    @property
    def standard(self) -> TrigAPlockStandard:
        """Return the ``TRIG:A:PLOCK:STANDARD`` command.

        **Description:**
            - This command sets or queries the communication standard used by the pattern lock
              trigger.

        **Usage:**
            - Using the ``.query()`` method will send the ``TRIG:A:PLOCK:STANDARD?`` query.
            - Using the ``.verify(value)`` method will send the ``TRIG:A:PLOCK:STANDARD?`` query and
              raise an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``TRIG:A:PLOCK:STANDARD value``
              command.

        **SCPI Syntax:**

        ::

            - TRIG:A:PLOCK:STANDARD {<STANDARD>}
            - TRIG:A:PLOCK:STANDARD?

        **Info:**
            - ``STANDARD`` communication standard used by the pattern lock trigger. See
              ``TRIGGER:A:COMMUNICATION:STANDARD`` for a list of standards.
        """
        return self._standard


class TrigA(SCPICmdRead):
    """The ``TRIG:A`` command tree.

    **Usage:**
        - Using the ``.query()`` method will send the ``TRIG:A?`` query.
        - Using the ``.verify(value)`` method will send the ``TRIG:A?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.plock``: The ``TRIG:A:PLOCK`` command tree.
    """

    def __init__(self, device: Optional["PIDevice"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._plock = TrigAPlock(device, f"{self._cmd_syntax}:PLOCK")

    @property
    def plock(self) -> TrigAPlock:
        """Return the ``TRIG:A:PLOCK`` command tree.

        **Usage:**
            - Using the ``.query()`` method will send the ``TRIG:A:PLOCK?`` query.
            - Using the ``.verify(value)`` method will send the ``TRIG:A:PLOCK?`` query and raise an
              AssertionError if the returned value does not match ``value``.

        Sub-properties:
            - ``.standard``: The ``TRIG:A:PLOCK:STANDARD`` command.
        """
        return self._plock


class Trig(SCPICmdRead):
    """The ``TRIG`` command tree.

    **Usage:**
        - Using the ``.query()`` method will send the ``TRIG?`` query.
        - Using the ``.verify(value)`` method will send the ``TRIG?`` query and raise an
          AssertionError if the returned value does not match ``value``.

    Properties:
        - ``.a``: The ``TRIG:A`` command tree.
    """

    def __init__(self, device: Optional["PIDevice"] = None, cmd_syntax: str = "TRIG") -> None:
        super().__init__(device, cmd_syntax)
        self._a = TrigA(device, f"{self._cmd_syntax}:A")

    @property
    def a(self) -> TrigA:
        """Return the ``TRIG:A`` command tree.

        **Usage:**
            - Using the ``.query()`` method will send the ``TRIG:A?`` query.
            - Using the ``.verify(value)`` method will send the ``TRIG:A?`` query and raise an
              AssertionError if the returned value does not match ``value``.

        Sub-properties:
            - ``.plock``: The ``TRIG:A:PLOCK`` command tree.
        """
        return self._a



"""Module containing helpers for the tm_devices package that only use standard lib imports."""
import ipaddress
import re

####################################################################################################
# Private Constants
####################################################################################################
_HOSTNAME_REGEX = re.compile(
    r"^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*"
    r"([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$"
)


####################################################################################################
# Public Functions
####################################################################################################


def validate_address(address: str) -> bool:
    """Validate an IP address or hostname.

    Args:
        address: The address to validate.

    Returns:
        A boolean, if False, the address was an IP address. If True, the address was a hostname.

    Raises:
        ValueError: Indicating the address was not a valid IP address or hostname.
    """
    is_hostname = False
    try:
        _ = ipaddress.ip_address(address)
    except ValueError as error:
        if not _HOSTNAME_REGEX.match(address):
            msg = f"{address=} is not a valid IP address or hostname."
            raise ValueError(msg) from error
        is_hostname = True

    return is_hostname



"""SMU2400 package init file."""
from tm_devices.drivers.pi.source_measure_units.smu2400.smu2400 import SMU2400
from tm_devices.drivers.pi.source_measure_units.smu2400.smu2400_interactive import (
    SMU2400Interactive,
)
from tm_devices.drivers.pi.source_measure_units.smu2400.smu2400_standard import SMU2400Standard
from tm_devices.drivers.pi.source_measure_units.smu2400.smu2401 import SMU2401
from tm_devices.drivers.pi.source_measure_units.smu2400.smu2410 import SMU2410
from tm_devices.drivers.pi.source_measure_units.smu2400.smu2450 import SMU2450
from tm_devices.drivers.pi.source_measure_units.smu2400.smu2460 import SMU2460
from tm_devices.drivers.pi.source_measure_units.smu2400.smu2461 import SMU2461
from tm_devices.drivers.pi.source_measure_units.smu2400.smu2470 import SMU2470

__all__ = [
    "SMU2400Interactive",
    "SMU2400Standard",
    "SMU2400",
    "SMU2401",
    "SMU2410",
    "SMU2450",
    "SMU2460",
    "SMU2461",
    "SMU2470",
]



"""The data commands module.

These commands are used in the following models:
LPD6, MSO4, MSO5, MSO5B, MSO5LP, MSO6, MSO6B

THIS FILE IS AUTO-GENERATED, IT SHOULD NOT BE MANUALLY MODIFIED.

Please report an issue if one is found.

Commands and Queries:

::

    - DATa {INIT|SNAp}
    - DATa:ENCdg {ASCIi| RIBinary| RPBinary| FPBinary| SRIbinary| SRPbinary| SFPbinary}
    - DATa:ENCdg?
    - DATa:FRAMESTARt <NR1>
    - DATa:FRAMESTARt?
    - DATa:FRAMESTOP {<NR1>|MAX}
    - DATa:FRAMESTOP?
    - DATa:MODe {VECtor|PIXmap}
    - DATa:MODe?
    - DATa:RESample <NR1>
    - DATa:RESample?
    - DATa:SOUrce <wfm>[<,><wfm>]
    - DATa:SOUrce:AVAILable?
    - DATa:SOUrce?
    - DATa:STARt <NR1>
    - DATa:STARt?
    - DATa:STOP <NR1>
    - DATa:STOP?
    - DATa:WIDth <NR1>
    - DATa:WIDth?
    - DATa?
"""
from typing import Optional, TYPE_CHECKING

from .._helpers import SCPICmdRead, SCPICmdWrite

if TYPE_CHECKING:
    from tm_devices.drivers.pi.pi_device import PIDevice


class DataWidth(SCPICmdWrite, SCPICmdRead):
    """The ``DATa:WIDth`` command.

    **Description:**
        - This command specifies the width, in bytes per point, for waveform data transferred from
          the instrument via the CURVe? query. (This command is synonymous with
          ``WFMOutpre:BYT_Nr``.)

    **Usage:**
        - Using the ``.query()`` method will send the ``DATa:WIDth?`` query.
        - Using the ``.verify(value)`` method will send the ``DATa:WIDth?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``DATa:WIDth value`` command.

    **SCPI Syntax:**

    ::

        - DATa:WIDth <NR1>
        - DATa:WIDth?

    **Info:**
        - ``<NR1>`` is an integer that indicates the number of bytes per point for the outgoing
          waveform data when queried using the CURVe? command.
    """


class DataStop(SCPICmdWrite, SCPICmdRead):
    """The ``DATa:STOP`` command.

    **Description:**
        - This command sets or queries the last data point that will be transferred when using the
          CURVE? query. When using the CURVE command, ``DATa:STOP`` is ignored. This command allows
          for the transfer of partial waveforms to the controller. If <NR1> is greater than the
          record length, then data will be transferred up to the record length. If both
          ``DATa:STARt`` and ``DATa:STOP`` are greater than the record length, the last data point
          in the record is returned. ``DATa:STARt`` and ``DATa:STOP`` are order independent. When
          ``DATa:STOP`` is less than ``DATa:STARt``, the values will be swapped internally for the
          CURVE? query. If you always want to transfer complete waveforms, set ``DATa:STARt`` to 1
          and ``DATa:STOP`` to the maximum record length, or larger. Changes to the record length
          value are not automatically reflected in the ``DATa:STOP`` value. As record length is
          varied, the ``DATa:STOP`` value must be explicitly changed to ensure the entire record is
          transmitted. In other words, curve results will not automatically and correctly reflect
          increases in record length if the distance from ``DATa:STARt`` to ``DATa:STOP`` stays
          smaller than the increased record length.

    **Usage:**
        - Using the ``.query()`` method will send the ``DATa:STOP?`` query.
        - Using the ``.verify(value)`` method will send the ``DATa:STOP?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``DATa:STOP value`` command.

    **SCPI Syntax:**

    ::

        - DATa:STOP <NR1>
        - DATa:STOP?

    **Info:**
        - ``<NR1>`` is the last data point that will be transferred, which ranges from 1 to the
          record length.
    """


class DataStart(SCPICmdWrite, SCPICmdRead):
    """The ``DATa:STARt`` command.

    **Description:**
        - This command sets or queries the starting data point for waveform transfer. This command
          allows for the transfer of partial waveforms to and from the instrument. Data will be
          transferred from <NR1> to ``DATa:STOP`` or the record length, whichever is less. If <NR1>
          is greater than the record length, the last data point in the record is transferred.
          ``DATa:STARt`` and ``DATa:STOP`` are order independent. When ``DATa:STOP`` is greater than
          ``DATa:STARt``, the values will be swapped internally for the CURVE? query.

    **Usage:**
        - Using the ``.query()`` method will send the ``DATa:STARt?`` query.
        - Using the ``.verify(value)`` method will send the ``DATa:STARt?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``DATa:STARt value`` command.

    **SCPI Syntax:**

    ::

        - DATa:STARt <NR1>
        - DATa:STARt?

    **Info:**
        - ``<NR1>`` is the first data point that will be transferred, which ranges from 1 to the
          record length.
    """


class DataSourceAvailable(SCPICmdRead):
    """The ``DATa:SOUrce:AVAILable`` command.

    **Description:**
        - This query returns a list of enumerations representing the source waveforms that are
          currently available for ``:CURVe?`` queries. This means that the waveforms have been
          acquired. If there are none, NONE is returned.

    **Usage:**
        - Using the ``.query()`` method will send the ``DATa:SOUrce:AVAILable?`` query.
        - Using the ``.verify(value)`` method will send the ``DATa:SOUrce:AVAILable?`` query and
          raise an AssertionError if the returned value does not match ``value``.

    **SCPI Syntax:**

    ::

        - DATa:SOUrce:AVAILable?
    """


class DataSource(SCPICmdWrite, SCPICmdRead):
    """The ``DATa:SOUrce`` command.

    **Description:**
        - This command sets or queries the location of waveform data that is transferred from the
          instrument by the CURVE Query.

    **Usage:**
        - Using the ``.query()`` method will send the ``DATa:SOUrce?`` query.
        - Using the ``.verify(value)`` method will send the ``DATa:SOUrce?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``DATa:SOUrce value`` command.

    **SCPI Syntax:**

    ::

        - DATa:SOUrce <wfm>[<,><wfm>]
        - DATa:SOUrce?

    **Info:**
        - ``<wfm>`` is the location of the waveform data that will be transferred from the
          instrument to the controller. It can consist of CH<x>, MATH<x>, REF<x>, DIGITALALL. Note
          that digital data is transferred as 16-bit data, with the least-significant bit
          representing D0, and the most-significant bit representing D15.
        - ``<wfm>`` can consist of the following.
        - ``CH<x>`` selects the specified analog channel as the source.
        - ``MATH<x>`` selects the specified reference waveform as the source. The reference number
          is specified by x, which ranges from 1 through 4.
        - ``REF<x>`` selects the specified reference waveform as the source. The reference number is
          specified by x, which ranges from 1 through 8.
        - ``CH<x>_D<x>`` selects the specified digital channel.
        - ``CH<x>_DAll`` selects the specified channel group of digital channels.
        - ``DIGITALALL`` selects digital waveforms as the source. The Digital data is transferred as
          16-bit data, with the least-significant bit representing D0, and the most-significant bit
          representing D15. The LSB always contains D0-D7 and MSB always contains D8-D15 data.
        - ``CH<x>_SV_NORMal`` , ``CH<x>_SV_AVErage``, ``CH<x>_SV_MAXHold``, ``CH<x>_SV_MINHold``
          selects the specified Spectrum View waveform.
        - ``CH<x>_MAG_VS_TIME`` , ``CH<x>_FREQ_VS_TIME``, ``CH<x>_PHASE_VS_TIME`` selects the
          specified RF vs. Time waveform.
        - ``CH<x>_SV_BASEBAND_IQ`` selects the specified RF baseband IQ data.

    Properties:
        - ``.available``: The ``DATa:SOUrce:AVAILable`` command.
    """

    def __init__(self, device: Optional["PIDevice"], cmd_syntax: str) -> None:
        super().__init__(device, cmd_syntax)
        self._available = DataSourceAvailable(device, f"{self._cmd_syntax}:AVAILable")

    @property
    def available(self) -> DataSourceAvailable:
        """Return the ``DATa:SOUrce:AVAILable`` command.

        **Description:**
            - This query returns a list of enumerations representing the source waveforms that are
              currently available for ``:CURVe?`` queries. This means that the waveforms have been
              acquired. If there are none, NONE is returned.

        **Usage:**
            - Using the ``.query()`` method will send the ``DATa:SOUrce:AVAILable?`` query.
            - Using the ``.verify(value)`` method will send the ``DATa:SOUrce:AVAILable?`` query and
              raise an AssertionError if the returned value does not match ``value``.

        **SCPI Syntax:**

        ::

            - DATa:SOUrce:AVAILable?
        """
        return self._available


class DataResample(SCPICmdWrite, SCPICmdRead):
    """The ``DATa:RESample`` command.

    **Description:**
        - This command sets or queries the resampling of outgoing waveform data. This command is
          equivalent to setting ``WFMOutpre:RESample``. Setting the ``DATa:RESample`` value causes
          the corresponding WFMOutpre value to be updated and vice versa.

    **Usage:**
        - Using the ``.query()`` method will send the ``DATa:RESample?`` query.
        - Using the ``.verify(value)`` method will send the ``DATa:RESample?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``DATa:RESample value`` command.

    **SCPI Syntax:**

    ::

        - DATa:RESample <NR1>
        - DATa:RESample?

    **Info:**
        - ``<NR1>`` is the resampling rate. The default value is 1, which means every sample is
          returned. A value of 2 returns every other sample, while a value of 3 returns every third
          sample, and so on.
    """


class DataMode(SCPICmdWrite, SCPICmdRead):
    """The ``DATa:MODe`` command.

    **Description:**
        - This command sets or queries the mode for waveform data sent to returned by CURVe?. When
          FastAcq mode is ON, and the value is PIXmap, it returns Fast Acquisition pixmap data or
          the vector data is returned. When the data mode is set as VECtor then you get the waveform
          sampled data. The Data width is reset to 1 or 2 instead of 4.

    **Usage:**
        - Using the ``.query()`` method will send the ``DATa:MODe?`` query.
        - Using the ``.verify(value)`` method will send the ``DATa:MODe?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``DATa:MODe value`` command.

    **SCPI Syntax:**

    ::

        - DATa:MODe {VECtor|PIXmap}
        - DATa:MODe?

    **Info:**
        - ``VECtor`` sets the mode for waveform data to vector.
        - ``PIXmap`` sets the mode for waveform data to pixmap.
    """


class DataFramestop(SCPICmdWrite, SCPICmdRead):
    """The ``DATa:FRAMESTOP`` command.

    **Description:**
        - This command sets or queries the last acquisition for waveform transfer using the CURVE?
          query. This is only relevant when History or FastFrame acquisition modes are enabled.

    **Usage:**
        - Using the ``.query()`` method will send the ``DATa:FRAMESTOP?`` query.
        - Using the ``.verify(value)`` method will send the ``DATa:FRAMESTOP?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``DATa:FRAMESTOP value`` command.

    **SCPI Syntax:**

    ::

        - DATa:FRAMESTOP {<NR1>|MAX}
        - DATa:FRAMESTOP?

    **Info:**
        - ``<NR1>`` is the last acquisition that will be transferred, which ranges from 1 to the
          number of History or FastFrame acquisitions. Results will be transferred from acquisitions
          ``DATa:FRAMESTARt`` to <NR1>. If <NR1> is greater than the number of acquisitions, then
          data will be transferred up to the last acquisition. If ``DATa:FRAMESTOP`` is less than
          ``DATa:FRAMESTARt``, then only a single acquisition at ``DATa:FRAMESTARt`` is transferred.
        - ``MAX`` indicates that data is always transferred up to the last acquisition.
    """


class DataFramestart(SCPICmdWrite, SCPICmdRead):
    """The ``DATa:FRAMESTARt`` command.

    **Description:**
        - This command sets or queries the starting acquisition for waveform transfer using the
          CURVE? query. This is only relevant when History or FastFrame acquisition modes are
          enabled.

    **Usage:**
        - Using the ``.query()`` method will send the ``DATa:FRAMESTARt?`` query.
        - Using the ``.verify(value)`` method will send the ``DATa:FRAMESTARt?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``DATa:FRAMESTARt value`` command.

    **SCPI Syntax:**

    ::

        - DATa:FRAMESTARt <NR1>
        - DATa:FRAMESTARt?

    **Info:**
        - ``<NR1>`` is the first acquisition that will be transferred, which ranges from 1 to the
          number of History or FastFrame acquisitions. Results are transferred from acquisition
          <NR1> to ``DATa:FRAMESTOP`` or the total number of acquisitions, whichever is less. If
          <NR1> is greater than the number of acquisitions, then only the last acquisition is
          transferred. If ``DATa:FRAMESTARt`` is greater than ``DATa:FRAMESTOP``, then only a single
          acquisition at <NR1> is transferred.
    """


class DataEncdg(SCPICmdWrite, SCPICmdRead):
    """The ``DATa:ENCdg`` command.

    **Description:**
        - This command sets or queries the format of outgoing waveform data. This command is
          equivalent to setting ``WFMOUTPRE:ENCDG``, ``WFMOUTPRE:BN_FMT``, and ``WFMOUTPRE:BYT_OR``.
          Setting the ``DATa:ENGdg`` value causes the corresponding WFMOutpre values to be updated
          and vice versa.

    **Usage:**
        - Using the ``.query()`` method will send the ``DATa:ENCdg?`` query.
        - Using the ``.verify(value)`` method will send the ``DATa:ENCdg?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``DATa:ENCdg value`` command.

    **SCPI Syntax:**

    ::

        - DATa:ENCdg {ASCIi| RIBinary| RPBinary| FPBinary| SRIbinary| SRPbinary| SFPbinary}
        - DATa:ENCdg?

    **Info:**
        - ``ASCIi`` specifies the ASCII representation of signed INT, FLOAT. If ASCII is the value,
          then ``:BN_Fmt`` and ``:BYT_Or`` are ignored. The following are the DATa and WFMOutpre
          parameter settings (separated by semicolons): ``:ENCdg`` = ASC ; ``:BN_Fmt`` = N/A ;
          ``:BYT_Or`` = N/A ; ``:BYT_NR`` = 1,2,4.
        - ``RIBinary`` specifies the positive integer data-point representation, with the most
          significant byte transferred first. When ``:BYT_Nr`` is 1, the range from 0 through 255.
          When ``:BYT_Nr`` is 2,the range is from 0 to 65,535. When ``:BYT_Nr`` is 4, then the
          waveform being queried would return Fast Acquisition Pixmap data (if fast acq is turned on
          and data mode is set to pixmap). The following are the DATa and WFMOutpre parameter
          settings (separated by semicolons): ``:ENCdg`` = BIN ; ``:BN_Fmt`` = RI ; ``:BYT_Or`` =
          MSB ; ``:BYT_NR`` = 1,2.
        - ``RPBinary`` specifies the positive integer data-point representation, with the most
          significant byte transferred first. When ``:BYT_Nr`` is 1, the range from 0 through 255.
          When ``:BYT_Nr`` is 2, the range is from 0 to 65,535. The following are the DATa and
          WFMOutpre parameter settings (separated by semicolons): ``:ENCdg`` = ASC ; ``:BN_Fmt`` =
          RP ; ``:BYT_Or`` = MSB ; ``:BYT_NR`` = 1,2.
        - ``FPBinary`` specifies the floating point (width = 4) data. The range is from -3.4 × 1038
          to 3.4 × 1038. The center of the screen is 0. The upper limit is the top of the screen and
          the lower limit is the bottom of the screen. The FPBinary argument is only applicable to
          math waveforms or ref waveforms saved from math waveforms. The following are the DATa and
          WFMOutpre parameter settings (separated by semicolons): ``:ENCdg`` = BIN ; ``:BN_Fmt`` =
          FP ; ``:BYT_Or`` = MSB ; ``:BYT_NR`` = 4.
        - ``SRIbinary`` is the same as RIBinary except that the byte order is swapped, meaning that
          the least significant byte is transferred first. This format is useful when transferring
          data to IBM compatible PCs. The following are the DATa and WFMOutpre parameter settings
          (separated by semicolons): ``:ENCdg`` = BIN ; ``:BN_Fmt`` = RI ; ``:BYT_Or`` = LSB ;
          ``:BYT_NR`` = 1,2.
        - ``SRPbinary`` is the same as RPBinary except that the byte order is swapped, meaning that
          the least significant byte is transferred first. This format is useful when transferring
          data to PCs. The following are the DATa and WFMOutpre parameter settings (separated by
          semicolons): ``:ENCdg`` = BIN ; ``:BN_Fmt`` = RP ; ``:BYT_Or`` = LSB ; ``:BYT_NR`` = 1,2.
        - ``SFPbinary`` specifies floating point data in IBM PC format. The SFPbinary argument only
          works on math waveforms or ref waveforms saved from math waveforms. The following are the
          DATa and WFMOutpre parameter settings (separated by semicolons): ``:ENCdg`` = BIN ;
          ``:BN_Fmt`` = FP ; ``:BYT_Or`` = LSB ; ``:BYT_NR`` = 4.
    """


#  pylint: disable=too-many-instance-attributes
class Data(SCPICmdWrite, SCPICmdRead):
    """The ``DATa`` command.

    **Description:**
        - This command sets or queries the format and location of the waveform data that is
          transferred with the CURVE command.

    **Usage:**
        - Using the ``.query()`` method will send the ``DATa?`` query.
        - Using the ``.verify(value)`` method will send the ``DATa?`` query and raise an
          AssertionError if the returned value does not match ``value``.
        - Using the ``.write(value)`` method will send the ``DATa value`` command.

    **SCPI Syntax:**

    ::

        - DATa {INIT|SNAp}
        - DATa?

    **Info:**
        - ``INIT`` initializes the waveform data parameters to their factory defaults except for
          ``DATa:STOP``, which isset to the current acquisition record length.
        - ``SNAp`` Sets ``DATa:STARt`` and ``DATa:STOP`` to match the current waveform cursor
          positions of WAVEVIEW1 CURSOR1 if these waveform cursors are currently on. If these
          waveform cursors are not on when the ``DATa SNAp`` command is sent, it is silently ignored
          and ``DATa:STARt`` and ``:STOP`` remain unchanged.

    Properties:
        - ``.encdg``: The ``DATa:ENCdg`` command.
        - ``.framestart``: The ``DATa:FRAMESTARt`` command.
        - ``.framestop``: The ``DATa:FRAMESTOP`` command.
        - ``.mode``: The ``DATa:MODe`` command.
        - ``.resample``: The ``DATa:RESample`` command.
        - ``.source``: The ``DATa:SOUrce`` command.
        - ``.start``: The ``DATa:STARt`` command.
        - ``.stop``: The ``DATa:STOP`` command.
        - ``.width``: The ``DATa:WIDth`` command.
    """

    def __init__(self, device: Optional["PIDevice"] = None, cmd_syntax: str = "DATa") -> None:
        super().__init__(device, cmd_syntax)
        self._encdg = DataEncdg(device, f"{self._cmd_syntax}:ENCdg")
        self._framestart = DataFramestart(device, f"{self._cmd_syntax}:FRAMESTARt")
        self._framestop = DataFramestop(device, f"{self._cmd_syntax}:FRAMESTOP")
        self._mode = DataMode(device, f"{self._cmd_syntax}:MODe")
        self._resample = DataResample(device, f"{self._cmd_syntax}:RESample")
        self._source = DataSource(device, f"{self._cmd_syntax}:SOUrce")
        self._start = DataStart(device, f"{self._cmd_syntax}:STARt")
        self._stop = DataStop(device, f"{self._cmd_syntax}:STOP")
        self._width = DataWidth(device, f"{self._cmd_syntax}:WIDth")

    @property
    def encdg(self) -> DataEncdg:
        """Return the ``DATa:ENCdg`` command.

        **Description:**
            - This command sets or queries the format of outgoing waveform data. This command is
              equivalent to setting ``WFMOUTPRE:ENCDG``, ``WFMOUTPRE:BN_FMT``, and
              ``WFMOUTPRE:BYT_OR``. Setting the ``DATa:ENGdg`` value causes the corresponding
              WFMOutpre values to be updated and vice versa.

        **Usage:**
            - Using the ``.query()`` method will send the ``DATa:ENCdg?`` query.
            - Using the ``.verify(value)`` method will send the ``DATa:ENCdg?`` query and raise an
              AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``DATa:ENCdg value`` command.

        **SCPI Syntax:**

        ::

            - DATa:ENCdg {ASCIi| RIBinary| RPBinary| FPBinary| SRIbinary| SRPbinary| SFPbinary}
            - DATa:ENCdg?

        **Info:**
            - ``ASCIi`` specifies the ASCII representation of signed INT, FLOAT. If ASCII is the
              value, then ``:BN_Fmt`` and ``:BYT_Or`` are ignored. The following are the DATa and
              WFMOutpre parameter settings (separated by semicolons): ``:ENCdg`` = ASC ; ``:BN_Fmt``
              = N/A ; ``:BYT_Or`` = N/A ; ``:BYT_NR`` = 1,2,4.
            - ``RIBinary`` specifies the positive integer data-point representation, with the most
              significant byte transferred first. When ``:BYT_Nr`` is 1, the range from 0 through
              255. When ``:BYT_Nr`` is 2,the range is from 0 to 65,535. When ``:BYT_Nr`` is 4, then
              the waveform being queried would return Fast Acquisition Pixmap data (if fast acq is
              turned on and data mode is set to pixmap). The following are the DATa and WFMOutpre
              parameter settings (separated by semicolons): ``:ENCdg`` = BIN ; ``:BN_Fmt`` = RI ;
              ``:BYT_Or`` = MSB ; ``:BYT_NR`` = 1,2.
            - ``RPBinary`` specifies the positive integer data-point representation, with the most
              significant byte transferred first. When ``:BYT_Nr`` is 1, the range from 0 through
              255. When ``:BYT_Nr`` is 2, the range is from 0 to 65,535. The following are the DATa
              and WFMOutpre parameter settings (separated by semicolons): ``:ENCdg`` = ASC ;
              ``:BN_Fmt`` = RP ; ``:BYT_Or`` = MSB ; ``:BYT_NR`` = 1,2.
            - ``FPBinary`` specifies the floating point (width = 4) data. The range is from -3.4 ×
              1038 to 3.4 × 1038. The center of the screen is 0. The upper limit is the top of the
              screen and the lower limit is the bottom of the screen. The FPBinary argument is only
              applicable to math waveforms or ref waveforms saved from math waveforms. The following
              are the DATa and WFMOutpre parameter settings (separated by semicolons): ``:ENCdg`` =
              BIN ; ``:BN_Fmt`` = FP ; ``:BYT_Or`` = MSB ; ``:BYT_NR`` = 4.
            - ``SRIbinary`` is the same as RIBinary except that the byte order is swapped, meaning
              that the least significant byte is transferred first. This format is useful when
              transferring data to IBM compatible PCs. The following are the DATa and WFMOutpre
              parameter settings (separated by semicolons): ``:ENCdg`` = BIN ; ``:BN_Fmt`` = RI ;
              ``:BYT_Or`` = LSB ; ``:BYT_NR`` = 1,2.
            - ``SRPbinary`` is the same as RPBinary except that the byte order is swapped, meaning
              that the least significant byte is transferred first. This format is useful when
              transferring data to PCs. The following are the DATa and WFMOutpre parameter settings
              (separated by semicolons): ``:ENCdg`` = BIN ; ``:BN_Fmt`` = RP ; ``:BYT_Or`` = LSB ;
              ``:BYT_NR`` = 1,2.
            - ``SFPbinary`` specifies floating point data in IBM PC format. The SFPbinary argument
              only works on math waveforms or ref waveforms saved from math waveforms. The following
              are the DATa and WFMOutpre parameter settings (separated by semicolons): ``:ENCdg`` =
              BIN ; ``:BN_Fmt`` = FP ; ``:BYT_Or`` = LSB ; ``:BYT_NR`` = 4.
        """
        return self._encdg

    @property
    def framestart(self) -> DataFramestart:
        """Return the ``DATa:FRAMESTARt`` command.

        **Description:**
            - This command sets or queries the starting acquisition for waveform transfer using the
              CURVE? query. This is only relevant when History or FastFrame acquisition modes are
              enabled.

        **Usage:**
            - Using the ``.query()`` method will send the ``DATa:FRAMESTARt?`` query.
            - Using the ``.verify(value)`` method will send the ``DATa:FRAMESTARt?`` query and raise
              an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``DATa:FRAMESTARt value`` command.

        **SCPI Syntax:**

        ::

            - DATa:FRAMESTARt <NR1>
            - DATa:FRAMESTARt?

        **Info:**
            - ``<NR1>`` is the first acquisition that will be transferred, which ranges from 1 to
              the number of History or FastFrame acquisitions. Results are transferred from
              acquisition <NR1> to ``DATa:FRAMESTOP`` or the total number of acquisitions, whichever
              is less. If <NR1> is greater than the number of acquisitions, then only the last
              acquisition is transferred. If ``DATa:FRAMESTARt`` is greater than ``DATa:FRAMESTOP``,
              then only a single acquisition at <NR1> is transferred.
        """
        return self._framestart

    @property
    def framestop(self) -> DataFramestop:
        """Return the ``DATa:FRAMESTOP`` command.

        **Description:**
            - This command sets or queries the last acquisition for waveform transfer using the
              CURVE? query. This is only relevant when History or FastFrame acquisition modes are
              enabled.

        **Usage:**
            - Using the ``.query()`` method will send the ``DATa:FRAMESTOP?`` query.
            - Using the ``.verify(value)`` method will send the ``DATa:FRAMESTOP?`` query and raise
              an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``DATa:FRAMESTOP value`` command.

        **SCPI Syntax:**

        ::

            - DATa:FRAMESTOP {<NR1>|MAX}
            - DATa:FRAMESTOP?

        **Info:**
            - ``<NR1>`` is the last acquisition that will be transferred, which ranges from 1 to the
              number of History or FastFrame acquisitions. Results will be transferred from
              acquisitions ``DATa:FRAMESTARt`` to <NR1>. If <NR1> is greater than the number of
              acquisitions, then data will be transferred up to the last acquisition. If
              ``DATa:FRAMESTOP`` is less than ``DATa:FRAMESTARt``, then only a single acquisition at
              ``DATa:FRAMESTARt`` is transferred.
            - ``MAX`` indicates that data is always transferred up to the last acquisition.
        """
        return self._framestop

    @property
    def mode(self) -> DataMode:
        """Return the ``DATa:MODe`` command.

        **Description:**
            - This command sets or queries the mode for waveform data sent to returned by CURVe?.
              When FastAcq mode is ON, and the value is PIXmap, it returns Fast Acquisition pixmap
              data or the vector data is returned. When the data mode is set as VECtor then you get
              the waveform sampled data. The Data width is reset to 1 or 2 instead of 4.

        **Usage:**
            - Using the ``.query()`` method will send the ``DATa:MODe?`` query.
            - Using the ``.verify(value)`` method will send the ``DATa:MODe?`` query and raise an
              AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``DATa:MODe value`` command.

        **SCPI Syntax:**

        ::

            - DATa:MODe {VECtor|PIXmap}
            - DATa:MODe?

        **Info:**
            - ``VECtor`` sets the mode for waveform data to vector.
            - ``PIXmap`` sets the mode for waveform data to pixmap.
        """
        return self._mode

    @property
    def resample(self) -> DataResample:
        """Return the ``DATa:RESample`` command.

        **Description:**
            - This command sets or queries the resampling of outgoing waveform data. This command is
              equivalent to setting ``WFMOutpre:RESample``. Setting the ``DATa:RESample`` value
              causes the corresponding WFMOutpre value to be updated and vice versa.

        **Usage:**
            - Using the ``.query()`` method will send the ``DATa:RESample?`` query.
            - Using the ``.verify(value)`` method will send the ``DATa:RESample?`` query and raise
              an AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``DATa:RESample value`` command.

        **SCPI Syntax:**

        ::

            - DATa:RESample <NR1>
            - DATa:RESample?

        **Info:**
            - ``<NR1>`` is the resampling rate. The default value is 1, which means every sample is
              returned. A value of 2 returns every other sample, while a value of 3 returns every
              third sample, and so on.
        """
        return self._resample

    @property
    def source(self) -> DataSource:
        """Return the ``DATa:SOUrce`` command.

        **Description:**
            - This command sets or queries the location of waveform data that is transferred from
              the instrument by the CURVE Query.

        **Usage:**
            - Using the ``.query()`` method will send the ``DATa:SOUrce?`` query.
            - Using the ``.verify(value)`` method will send the ``DATa:SOUrce?`` query and raise an
              AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``DATa:SOUrce value`` command.

        **SCPI Syntax:**

        ::

            - DATa:SOUrce <wfm>[<,><wfm>]
            - DATa:SOUrce?

        **Info:**
            - ``<wfm>`` is the location of the waveform data that will be transferred from the
              instrument to the controller. It can consist of CH<x>, MATH<x>, REF<x>, DIGITALALL.
              Note that digital data is transferred as 16-bit data, with the least-significant bit
              representing D0, and the most-significant bit representing D15.
            - ``<wfm>`` can consist of the following.
            - ``CH<x>`` selects the specified analog channel as the source.
            - ``MATH<x>`` selects the specified reference waveform as the source. The reference
              number is specified by x, which ranges from 1 through 4.
            - ``REF<x>`` selects the specified reference waveform as the source. The reference
              number is specified by x, which ranges from 1 through 8.
            - ``CH<x>_D<x>`` selects the specified digital channel.
            - ``CH<x>_DAll`` selects the specified channel group of digital channels.
            - ``DIGITALALL`` selects digital waveforms as the source. The Digital data is
              transferred as 16-bit data, with the least-significant bit representing D0, and the
              most-significant bit representing D15. The LSB always contains D0-D7 and MSB always
              contains D8-D15 data.
            - ``CH<x>_SV_NORMal`` , ``CH<x>_SV_AVErage``, ``CH<x>_SV_MAXHold``, ``CH<x>_SV_MINHold``
              selects the specified Spectrum View waveform.
            - ``CH<x>_MAG_VS_TIME`` , ``CH<x>_FREQ_VS_TIME``, ``CH<x>_PHASE_VS_TIME`` selects the
              specified RF vs. Time waveform.
            - ``CH<x>_SV_BASEBAND_IQ`` selects the specified RF baseband IQ data.

        Sub-properties:
            - ``.available``: The ``DATa:SOUrce:AVAILable`` command.
        """
        return self._source

    @property
    def start(self) -> DataStart:
        """Return the ``DATa:STARt`` command.

        **Description:**
            - This command sets or queries the starting data point for waveform transfer. This
              command allows for the transfer of partial waveforms to and from the instrument. Data
              will be transferred from <NR1> to ``DATa:STOP`` or the record length, whichever is
              less. If <NR1> is greater than the record length, the last data point in the record is
              transferred. ``DATa:STARt`` and ``DATa:STOP`` are order independent. When
              ``DATa:STOP`` is greater than ``DATa:STARt``, the values will be swapped internally
              for the CURVE? query.

        **Usage:**
            - Using the ``.query()`` method will send the ``DATa:STARt?`` query.
            - Using the ``.verify(value)`` method will send the ``DATa:STARt?`` query and raise an
              AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``DATa:STARt value`` command.

        **SCPI Syntax:**

        ::

            - DATa:STARt <NR1>
            - DATa:STARt?

        **Info:**
            - ``<NR1>`` is the first data point that will be transferred, which ranges from 1 to the
              record length.
        """
        return self._start

    @property
    def stop(self) -> DataStop:
        """Return the ``DATa:STOP`` command.

        **Description:**
            - This command sets or queries the last data point that will be transferred when using
              the CURVE? query. When using the CURVE command, ``DATa:STOP`` is ignored. This command
              allows for the transfer of partial waveforms to the controller. If <NR1> is greater
              than the record length, then data will be transferred up to the record length. If both
              ``DATa:STARt`` and ``DATa:STOP`` are greater than the record length, the last data
              point in the record is returned. ``DATa:STARt`` and ``DATa:STOP`` are order
              independent. When ``DATa:STOP`` is less than ``DATa:STARt``, the values will be
              swapped internally for the CURVE? query. If you always want to transfer complete
              waveforms, set ``DATa:STARt`` to 1 and ``DATa:STOP`` to the maximum record length, or
              larger. Changes to the record length value are not automatically reflected in the
              ``DATa:STOP`` value. As record length is varied, the ``DATa:STOP`` value must be
              explicitly changed to ensure the entire record is transmitted. In other words, curve
              results will not automatically and correctly reflect increases in record length if the
              distance from ``DATa:STARt`` to ``DATa:STOP`` stays smaller than the increased record
              length.

        **Usage:**
            - Using the ``.query()`` method will send the ``DATa:STOP?`` query.
            - Using the ``.verify(value)`` method will send the ``DATa:STOP?`` query and raise an
              AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``DATa:STOP value`` command.

        **SCPI Syntax:**

        ::

            - DATa:STOP <NR1>
            - DATa:STOP?

        **Info:**
            - ``<NR1>`` is the last data point that will be transferred, which ranges from 1 to the
              record length.
        """
        return self._stop

    @property
    def width(self) -> DataWidth:
        """Return the ``DATa:WIDth`` command.

        **Description:**
            - This command specifies the width, in bytes per point, for waveform data transferred
              from the instrument via the CURVe? query. (This command is synonymous with
              ``WFMOutpre:BYT_Nr``.)

        **Usage:**
            - Using the ``.query()`` method will send the ``DATa:WIDth?`` query.
            - Using the ``.verify(value)`` method will send the ``DATa:WIDth?`` query and raise an
              AssertionError if the returned value does not match ``value``.
            - Using the ``.write(value)`` method will send the ``DATa:WIDth value`` command.

        **SCPI Syntax:**

        ::

            - DATa:WIDth <NR1>
            - DATa:WIDth?

        **Info:**
            - ``<NR1>`` is an integer that indicates the number of bytes per point for the outgoing
              waveform data when queried using the CURVe? command.
        """
        return self._width



"""Test & Measurement Device Management.

Provides access to commonly imported items from the tm_devices package.

Examples:
    >>> from tm_devices import DeviceManager
    >>> with DeviceManager() as dm:  # doctest: +SKIP
    ...     print(dm.is_open)
    2023-09-21 01:50:16.031 - Opening DeviceManager
    True
    2023-09-21 01:50:16.031 - DeviceManager Closed
"""
from importlib.metadata import version

from tm_devices.device_manager import DeviceManager, print_available_visa_devices
from tm_devices.helpers.constants_and_dataclasses import (
    PACKAGE_NAME,
    PYVISA_PY_BACKEND,
    SYSTEM_DEFAULT_VISA_BACKEND,
)
from tm_devices.helpers.enums import SupportedModels

# Read version from installed package.
__version__ = version(PACKAGE_NAME)

__all__ = [
    "DeviceManager",
    "print_available_visa_devices",
    "PYVISA_PY_BACKEND",
    "SupportedModels",
    "SYSTEM_DEFAULT_VISA_BACKEND",
]
